#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 18:24:20 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.6.4'
__release__ = 20240811
__all__ = ['unyts_parameters_', 'print_path', 'reload', 'raise_error', 'cache', 'set_density', 'get_density',
           'recursion_limit', 'verbose', 'set_algorithm']

import logging
import os.path
from json import load as json_load, dump as json_dump
from os.path import isfile, isdir
from pathlib import Path
from sys import getrecursionlimit

ini_path = Path(__file__).with_name('parameters.ini').absolute()
ini_backup = Path(__file__).with_name('parameters.backup').absolute()
dir_path = os.path.dirname(ini_path) + '/'
off_switches = ('off', 'no', 'not', '')
__max_recursion_default__ = 12
__max_generations_default__ = 25

class UnytsParameters(object):
    """
    class to load the user preferences
    """

    def __init__(self, reload=None) -> None:
        self.config_files_folder_ = dir_path
        self.print_path_ = False
        self.cache_ = True
        self.raise_error_ = True
        self.verbose_ = False
        self.verbose_details_ = 0
        self.reduce_parentheses_ = True
        self.show_version_ = True
        self.density_ = 1.0  # g/cm3
        self.fvf_ = 1.0  # res_vol/std_vol
        self.max_recursion_ = __max_recursion_default__
        self.algorithm_ = 'lean_BFS'
        self.max_generations_ = __max_generations_default__
        self.load_params()
        self.reload_ = self.reload_ if reload is None else bool(reload)
        self.memory_ = not self.reload_
        self.last_path_str = ""
        self.gui = False

    def load_params(self) -> None:
        if isfile(ini_path):
            with open(ini_path, 'r') as f:
                params = json_load(f)
        else:
            params = {'print_path': False,
                      'cache': True,
                      'reload': True,
                      'raise_error': True,
                      'verbose': False,
                      'verbose_details': 0,
                      'reduce_parentheses': True,
                      'show_version': False,
                      'density': 1.0, # g/cm3
                      'max_recursion': __max_recursion_default__,
                      'algorithm': 'lean_BFS',
                      'max_generations': __max_generations_default__,
                      'fvf': 1.0,
                      'config_files_folder': None}
            with open(ini_path, 'w') as f:
                json_dump(params, f)
        self.print_path_ = params['print_path'] if 'print_path' in params else False
        self.cache_ = params['cache'] if 'cache' in params else True
        self.reload_ = params['reload'] if 'reload' in params else False
        self.raise_error_ = params['raise_error'] if 'raise_error' in params else True
        self.verbose_ = params['verbose'] if 'verbose' in params else False
        self.verbose_details_ = params['verbose_details'] if 'verbose_details' in params else 0
        self.reduce_parentheses_ = params['reduce_parentheses'] if 'reduce_parentheses' in params else True
        self.show_version_ = params['show_version'] if 'show_version' in params else True
        self.density_ = params['density'] if 'density' in params else 1.0  # g/cm3
        self.max_recursion_ = params['max_recursion'] if 'max_recursion' in params else __max_recursion_default__
        self.algorithm_ = params['algorithm'] if 'algorithm' in params else 'BFS'
        self.max_generations_ = params['max_generations'] if 'max_generations' in params else __max_generations_default__
        self.fvf_ = params['fvf'] if 'fvf' in params else 1.0
        self.config_files_folder_ = dir_path if ('config_files_folder' not in params or params['config_files_folder'] is None) \
            else params['config_files_folder'] if ('config_files_folder' in params and isdir(params['config_files_folder'])) \
            else self.config_files_folder_

        if self.show_version_ and isfile(ini_backup):
            with open(ini_backup, 'r') as f:
                params = json_load(f)
            logging.info("Restoring configuration from previous installation...")
            self.print_path_ = params['print_path'] if 'print_path' in params else False
            self.cache_ = params['cache'] if 'cache' in params else True
            self.raise_error_ = params['raise_error'] if 'raise_error' in params else True
            self.verbose_ = params['verbose'] if 'verbose' in params else False
            self.verbose_details_ = params['verbose_details'] if 'verbose_details' in params else 0
            self.reduce_parentheses_ = params['reduce_parentheses'] if 'reduce_parentheses' in params else True
            self.density_ = params['density'] if 'density' in params else 1.0  # g/cm3
            self.max_recursion_ = params['max_recursion'] if 'max_recursion' in params else __max_recursion_default__
            self.algorithm_ = params['algorithm'] if 'algorithm' in params else 'BFS'
            self.max_generations_ = params['max_generations'] if 'max_generations' in params else __max_generations_default__
            self.fvf_ = params['fvf'] if 'fvf' in params else 1.0
            self.config_files_folder_ = dir_path if ('config_files_folder' not in params or params['config_files_folder'] is None) \
                else params['config_files_folder'] if ('config_files_folder' in params and isdir(params['config_files_folder'])) \
                else self.config_files_folder_

    def save_params(self) -> None:
        params = {'print_path': self.print_path_,
                  'cache': self.cache_,
                  'reload': self.reload_,
                  'raise_error': self.raise_error_,
                  'verbose': self.verbose_,
                  'verbose_details': self.verbose_details_,
                  'reduce_parentheses': self.reduce_parentheses_,
                  'show_version': self.show_version_,
                  'density': self.density_,
                  'max_recursion': self.max_recursion_,
                  'algorithm': self.algorithm_,
                  'max_generations': self.max_generations_,
                  'fvf': self.fvf_,
                  'config_files_folder': self.config_files_folder_ if self.config_files_folder_ != dir_path else None}
        with open(ini_path, 'w') as f:
            json_dump(params, f)
        with open(ini_backup, 'w') as f:
            json_dump(params, f)

    def print_path(self, switch=None) -> None:
        if switch is None:
            self.print_path_ = not self.print_path_
        elif type(switch) is str:
            if switch.lower().strip() in off_switches:
                self.print_path_ = False
            else:
                self.print_path_ = True
        else:
            self.print_path_ = bool(switch)
        logging.info(f"print path {'ON' if self.print_path_ else 'OFF'}")
        self.save_params()

    def cache(self, switch=None) -> None:
        if switch is None:
            self.cache_ = not self.cache_
        elif type(switch) is str:
            if switch.lower().strip() in off_switches:
                self.cache_ = False
            else:
                self.cache_ = True
        else:
            self.cache_ = bool(switch)
        logging.info(f"cache {'ON' if self.cache_ else 'OFF'}")
        self.save_params()

    def raise_error(self, switch=None) -> None:
        if switch is None:
            self.raise_error_ = not self.raise_error_
        elif type(switch) is str:
            if switch.lower().strip() in off_switches:
                self.raise_error_ = False
            else:
                self.raise_error_ = True
        else:
            self.raise_error_ = bool(switch)
        logging.info(f"raise_error {'ON' if self.raise_error_ else 'OFF'}")
        self.save_params()

    def verbose(self, switch=None) -> None:
        if type(switch) is int:
            self.verbose_details_ = switch
            if switch <= 0:
                switch = False
            else:  # switch >= 1:
                switch = True
        if switch is None:
            self.verbose_ = not self.verbose_
        elif type(switch) is str:
            if switch.lower().strip() in off_switches:
                self.verbose_ = False
            else:
                self.verbose_ = True
        else:
            self.verbose_ = bool(switch)
        logging.info(f"verbose {'ON' if self.verbose_ else 'OFF'}")
        self.save_params()

    def reload_next_time(self, switch=None):
        if switch is None:
            self.reload_ = not self.reload_
        elif type(switch) is str:
            if switch.lower().strip() in off_switches:
                self.reload_ = False
            else:
                self.reload_ = True
        else:
            self.reload_ = bool(switch)
        if self.reload_:
            logging.info("Unyts will recreate dictionaries next time.")
        else:
            logging.info("Unyts will try to load dictionaries from cache next time.")
        self.save_params()

    def recursion_limit(self, limit=None):
        if limit is None:
            return self.max_recursion_
        elif (type(limit) is str and limit.lower() == 'max') or limit is False:
            self.max_recursion_ = getrecursionlimit() - 15
        elif limit is True:
            self.max_recursion_ = __max_recursion_default__
        elif type(limit) is int:
            self.max_recursion_ = min(limit, getrecursionlimit() - 15)
        else:
            raise ValueError(f"not valid recursion limit, must be an integer or boolean.")
        self.save_params()
        return self.max_recursion_

    def generations_limit(self, limit=None):
        if limit is None:
            return self.max_generations_
        elif (type(limit) is str and limit.lower() == 'max') or limit is False:
            self.max_generations_ = __max_generations_default__
        elif limit is True:
            self.max_generations_ = __max_generations_default__
        elif type(limit) is int:
            self.max_generations_ = limit
        else:
            raise ValueError(f"not valid generations limit, must be an integer or boolean.")
        self.save_params()
        return self.max_generations_

    def get_algorithm(self):
        return self.algorithm_

    def set_algorithm(self, algorithm:str):
        if algorithm not in ['BFS', 'lean_BFS', 'DFS']:
            logging.error(f"Valid algorithms are 'BFS', 'lean_BFS' and 'DFS' not {algorithm}.")
        else:
            self.algorithm_ = algorithm
            logging.info(f"{algorithm} set as search algorithm.")
            self.save_params()

    def get_algorithm(self):
        return self.algorithm_

    def set_user_folder(self, path=None):
        if path is None:
            self.config_files_folder_ = dir_path
        elif isdir(path):
            self.config_files_folder_ = path + '' if not path.endswith('/') and not path.endswith('\\') else '/'
        else:
            logging.warning(f"Folder {path} doesn't exists, user folder not changed.")

    def get_user_folder(self):
        return self.config_files_folder_


unyts_parameters_ = UnytsParameters()


def print_path(switch=None) -> None:
    unyts_parameters_.print_path(switch)


def raise_error(switch=None) -> None:
    unyts_parameters_.raise_error(switch)


def verbose(switch=None) -> None:
    unyts_parameters_.verbose(switch)


def cache(switch=None) -> None:
    unyts_parameters_.cache(switch)

def recursion_limit(limit=None) -> int:
    return unyts_parameters_.recursion_limit(limit)


def set_density(density: float = None, units: str = 'g/cm3') -> None:
    from .units.ratios import Density
    if density is None:
        print('Please enter density g/cm³ or enter a tuple density, units like: 997, kg/m³ :')
        while density is None:
            density = input(' density (g/cm³) = ')
            if ',' in density and len(density.split(',')) == 2:
                density, units = density.strip('()').split(',')
                density, units = float(density), units.strip(' ()')
            elif ',' not in density:
                try:
                    density = float(density)
                except:
                    density = None
            else:
                density = None
    if not isinstance(density, (int, float)) and not type(density) is Density:
        raise ValueError("'density' must be a float or int.")
    elif type(density) is Density:
        density = density.to('g/cm3')
        density, units = density.value, density.unit
    if units.lower() not in ('g/cc', 'g/cm3', 'g/cm³'):
        from .converter import convert
        from .errors import NoConversionFoundError
        try:
            density = convert(density, units, 'g/cm3')
        except NoConversionFoundError:
            raise ValueError("density 'units' are not valid.")
        if density is None:
            raise ValueError("density 'units' are not valid.")
    unyts_parameters_.density_ = density
    logging.info(f"density set to {density} g/cm³")
    unyts_parameters_.save_params()


def _get_density() -> float:
    while unyts_parameters_.density_ is None:
        density = input("Please set density value: ")
        try:
            density = float(density)
            unyts_parameters_.density_ = density
        except ValueError:
            print("'density' must be float or int.")
            density = None
    return unyts_parameters_.density_


def get_density():
    from .units.ratios import Density
    if unyts_parameters_.density_ is None:
        raise ValueError("'density' is not set.")
    return Density(unyts_parameters_.density_, 'g/cc')


def reload() -> None:
    unyts_parameters_.reload_ = True
    unyts_parameters_.save_params()
    logging.info("On next 'import unyts', the dictionary and network will be re-created.\n It might be required to restart the Python kernel.")


def set_algorithm(algorithm:str):
    if algorithm not in ['BFS', 'lean_BFS', 'DFS']:
        raise ValueError(f"valid algorithms are 'BFS', 'lean_BFS' and 'DFS' not {algorithm}.")
    unyts_parameters_.set_algorithm(algorithm)


def get_algorithm():
    return unyts_parameters_.algorithm_



# -*- coding: utf-8 -*-
"""
Created on Sat Jan 28 09:41:45 2023

@author: martin
"""
__version__ = '0.0.1'
__release__ = 20230128

from unyts import units, Unit

# Data
bit = units(1, 'b')
byte = units(1, 'B')

# Length
millimetre = units(1, 'mm')
millimeter = millimetre
centimeter = units(1, 'cm')
meter = units(1, 'm')
kilometre = units(1, 'km')
kilometer = kilometre
thou = units(1, 'th')
tenth = units(1, 'te')
inch = units(1, 'in')
foot = units(1, 'ft')
feet = foot
chain = units(1, 'ch')
yard = units(1, 'yd')
furlong = units(1, 'fur')
mile = units(1, 'mi')
league = units(1, 'lea')

# Area
acre = units(1, 'acre')

# Volume
barrel = units(1, 'stb')
stb = barrel
sm3 = units(1, 'sm3')
litre = units(1, 'l')
gill = units(1, 'gi')
pint = units(1, 'pt')
quart = units(1, 'qt')
gallonUS = units(1, 'galUS')
gallonUK = units(1, 'galUK')

# Weight
gram = units(1, 'g')
kilogram = units(1, 'kg')
grain = units(1, 'gr')
pennyweight = units(1, 'pwt')
dram = units(1, 'dr')
pound = units(1, 'lb')
quarter = units(1, 'qrt')
ton = units(1, 'long ton')

# ounce as Volume or Weight
ounce = Unit(1, 'oz')

# Time
second = units(1, 'sec')
minute = units(1, 'min')
hour = units(1, 'hour')
day = units(1, 'day')
week = units(1, 'week')
month = units(1, 'month')
year = units(1, 'year')

# Temperature
celsius = units(1, 'deg C')
fahrenheit = units(1, 'deg F')
rankine = units(1, 'deg R')
kelvin = units(1, 'deg K')

# Pressure
pascal = units(1, 'Pa')
kilopascal = units(1, 'kPa')
torr = units(1, 'mmHg')
mmgh = torr
psi = units(1, 'psia')
bar = units(1, 'barsa')
atmosphere = units(1, 'atm')

# Force
newton = units(1, 'N')
dyne = units(1, 'dyne')

# Energy
joule = units(1, 'J')
btu = units(1, 'btu')

# Power
horsepower = units(1, 'hp')
watt = units(1, 'W')

volt = units(1, 'V')
ampere = units(1, 'A')
ohm = units(1, 'Ω')
siemen = units(1, '℧')
farad = units(1, 'Farad')
coulomb = units(1, 'Q')
henry = units(1, 'H')

# Frequency
hertz = units(1, 'Hz')
rpm = units(1, 'RPM')



# -*- coding: utf-8 -*-
"""
Created on Sat Jan 21 22:34:17 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""
import logging

from unyts.dictionaries import dictionary, uncertain_names
from unyts import units
from string import ascii_uppercase


def key2name(txt):
    return (''.join([('_' if s in ascii_uppercase else '') + s.lower() for s in txt])).strip('_')


known_duplicated_units = ['F', 'l', 'ounces', 'OUNCES', 'ounce', 'OUNCE', 'oz', 'Ohm', 'ohm', 'OHM']
known_duplicated_keys = ['PressureGradient', 'Date', 'Impedance', 'Capacitance']

max_check = 25

for key in dictionary:
    if key in ['Impedance', 'Date', 'otherUnits']:
        continue
    # print(key)
    if len(dictionary[key]) > max_check:
        to_check = [u for u in dictionary[key] if
                    (u.count('*') <= 1 and u.count('/') == 0) or (u.count('/') <= 1 and u.count('*') == 0)]
    else:
        to_check = dictionary[key]
    for unit in to_check[:max_check]:
        print(key, unit)
        if key in ['Length', 'Rate'] and (len(unit) <= 2 and unit.endswith('l')) or (
                '/' in unit and len(unit.split('/')[0]) <= 2 and unit.split('/')[0].endswith('l')):
            continue  # litre is defines as length for SI sufixes (linear multiplier)
        if unit.lower() in ['ounce', 'oz', 'ounces'] and key == 'Weight':
            continue  # 'ounce' can be volume or mass. It is always instantiated as Volume but ounce can be converted to mass
        if key == 'Density' and unit in dictionary['PressureGradient']:
            continue  # pressure gradients have density units
        if key == 'Capacitance' and unit == 'F':
            continue  # F for Farad
        if unit not in known_duplicated_units and key not in known_duplicated_keys:
            rept_units = [k for k in dictionary for u in dictionary[k] if u == unit]
        else:
            rept_units = []
        if len(rept_units) > 1:
            logging.warning('unit ' + str(unit) + ' repeated in more than one dictionary:\n ' + '\n '.join(rept_units))
        if unit in uncertain_names:
            assert units(1, unit).name.lower() == 'unit'
        else:
            assert units(1, unit).name.lower() == key2name(key)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 21:34:21 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

from unyts import units, convert, set_fvf
from unyts.dictionaries import dictionary
from unyts.units.unitless import Dimensionless, Percentage
from unyts.operations import unit_product, unit_division
import numpy as np

num = 3
array = np.array([1, 2, 3, 4, 5])
limit_dict_units = 3

set_fvf(1.10)  # set FVF for these tests

for kind in [k for k in dictionary if k in ['Length']]:
    for unit1 in range(len(dictionary[kind][:limit_dict_units])-1):
        if len(dictionary[kind][unit1]) <= 2 and dictionary[kind][unit1].endswith('l'):
            continue
        print(dictionary[kind][unit1])
        u0 = units(2.78, dictionary[kind][unit1])
        u1 = units(3.0, dictionary[kind][unit1])

        assert str(u1) == (str(u1.value) + '_' + str(u1.unit))
        assert -u1 == (u1 * -1)
        assert bool(u1) is False if u1.kind in (Dimensionless, Percentage) else True
        assert abs(u1) == u1.kind(abs(u1.value), u1.unit)
        assert round(u0, 0) == u1

        for op in ('+', '-', '*', '/', '%'):
            assert eval('u1 ' + op + ' num') == units(eval('u1.value ' + op + ' num'), u1.unit)
        assert u1 ** 2 == units(u1.value ** 2, unit_product(u1.unit, u1.unit))

        for unit2 in range(unit1, len(dictionary[kind][:limit_dict_units])):
            if len(dictionary[kind][unit2]) <= 2 and dictionary[kind][unit2].endswith('l'):
                continue
            u2 = units(3, dictionary[kind][unit2])
            print(u1, '<->', u2)

            assert u1.to(u2) == units(convert(u1.value, u1.unit, u2.unit), u2.unit)

            assert (u1 + u2) == units(u1.value + convert(u2.value, u2.unit, u1.unit), u1.unit)
            assert (u2 + u1) == units(u2.value + convert(u1.value, u1.unit, u2.unit), u2.unit)

            assert (u1 - u2) == units(u1.value - convert(u2.value, u2.unit, u1.unit), u1.unit)
            assert (u2 - u1) == units(u2.value - convert(u1.value, u1.unit, u2.unit), u2.unit)

            assert (u1 * u2) == units(u1.value * convert(u2.value, u2.unit, u1.unit), unit_product(u1.unit, u2.unit))
            assert (u2 * u1) == units(u2.value * convert(u1.value, u1.unit, u2.unit), unit_product(u2.unit, u1.unit))

            assert (u1 / u2) == units(u1.value / convert(u2.value, u2.unit, u1.unit), unit_division(u1.unit, u2.unit))
            assert (u2 / u1) == units(u2.value / convert(u1.value, u1.unit, u2.unit), unit_division(u2.unit, u1.unit))

            assert (u1 // u2) == units(u1.value // convert(u2.value, u2.unit, u1.unit), unit_division(u1.unit, u2.unit))
            assert (u2 // u1) == units(u2.value // convert(u1.value, u1.unit, u2.unit), unit_division(u2.unit, u1.unit))

            assert (u1 < u2) == (u1.value < convert(u2.value, u2.unit, u1.unit))
            assert (u2 < u1) == (u2.value < convert(u1.value, u1.unit, u2.unit))

            assert (u1 <= u2) is (u1.value <= convert(u2.value, u2.unit, u1.unit))
            assert (u2 <= u1) is (u2.value <= convert(u1.value, u1.unit, u2.unit))

            assert (u1 == u2) is (u1.value == convert(u2.value, u2.unit, u1.unit))
            assert (u2 == u1) is (u2.value == convert(u1.value, u1.unit, u2.unit))

            assert (u1 != u2) is (u1.value != convert(u2.value, u2.unit, u1.unit))
            assert (u2 != u1) is (u2.value != convert(u1.value, u1.unit, u2.unit))

            assert (u1 >= u2) is (u1.value >= convert(u2.value, u2.unit, u1.unit))
            assert (u2 >= u1) is (u2.value >= convert(u1.value, u1.unit, u2.unit))

            assert (u1 > u2) is (u1.value > convert(u2.value, u2.unit, u1.unit))
            assert (u2 > u1) is (u2.value > convert(u1.value, u1.unit, u2.unit))


for kind in [k for k in dictionary if k in ['Length']]:  # [k for k in dictionary if k not in ['Dimensionless', 'Percentage']]:
    for unit1 in range(len(dictionary[kind][:limit_dict_units])-1):
        if len(dictionary[kind][unit1]) <= 2 and dictionary[kind][unit1].endswith('l'):
            continue
        print(dictionary[kind][unit1])
        u1 = units(array, dictionary[kind][unit1])

        assert str(u1) == (str(u1.value) + '_' + str(u1.unit))
        # assert (-u1).equals((u1 * -1))
        # # assert bool(u1) is False if u1.kind in (Dimensionless, Percentage) else True
        # assert abs(u1) == u1.kind(abs(u1.value), u1.unit)

        # for op in ('+', '-', '*', '/', '%'):
        #     assert eval('u1 ' + op + ' num') == units(eval('u1.value ' + op + ' num'), u1.unit)
        # assert u1 ** 2 == units(u1.value ** 2, unitProduct(u1.unit, u1.unit))

        # for unit2 in range(unit1, len(dictionary[kind][:limit_dict_units])):
        #     u2 = units(3, dictionary[kind][unit1])
        #     print(u1, '<->', u2)

        #     assert u1.to(u2) == convert(u1.value, u1.unit, u2.unit)

        #     assert (u1 + u2) == units(u1.value + convert(u2.value, u2.unit, u1.unit), u1.unit)
        #     assert (u2 + u1) == units(u2.value + convert(u1.value, u1.unit, u2.unit), u2.unit)

        #     assert (u1 - u2) == units(u1.value - convert(u2.value, u2.unit, u1.unit), u1.unit)
        #     assert (u2 - u1) == units(u2.value - convert(u1.value, u1.unit, u2.unit), u2.unit)

        #     assert (u1 * u2) == units(u1.value * convert(u2.value, u2.unit, u1.unit), unitProduct(u1.unit, u2.unit))
        #     assert (u2 * u1) == units(u2.value * convert(u1.value, u1.unit, u2.unit), unitProduct(u2.unit, u1.unit))

        #     assert (u1 / u2) == units(u1.value / convert(u2.value, u2.unit, u1.unit), unitDivision(u1.unit, u2.unit))
        #     assert (u2 / u1) == units(u2.value / convert(u1.value, u1.unit, u2.unit), unitDivision(u2.unit, u1.unit))

        #     assert (u1 // u2) == units(u1.value // convert(u2.value, u2.unit, u1.unit), unitDivision(u1.unit, u2.unit))
        #     assert (u2 // u1) == units(u2.value // convert(u1.value, u1.unit, u2.unit), unitDivision(u2.unit, u1.unit))

assert units(6, 'ft') / units(12, 'in') == units(6, 'ft/ft')
assert (units(6, 'ft') / units(12, 'in')).units == 'ft/ft'
assert units(6, 'ft') @ units(12, 'in') == units(0.5, 'ft/in')
assert (units(6, 'ft') @ units(12, 'in')).units == 'ft/in'

assert round((units(12, 'in')**2 * units(7, 'm')).to('m3'), 6) == round((units(7, 'm') * units(12, 'in')**2), 6)

assert units(0.0172, 'm').round(2).value == 0.02
assert units(0.0172, 'm').round(-1).value == 0.02
assert units(0.0172, 'm').round(-2).value == 0.017
assert units(1.0172, 'm').round(-1).value == 1
assert units(1.0172, 'm').round(-3).value == 1.02

assert units(1.01, 'm').equals(units(1.02, 'm'), 1) is True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.7'
__release__ = 20231230
__all__ = ['array_like', 'number', 'numeric',
           'array_like_', 'number_', 'numeric_',
           'unit_or_str', 'Empty', 'str_Empty']

from typing import Union
from ..unit_class import Unit
from ..Empty import Empty, str_Empty

unit_or_str = Union[str, Unit]

try:
    from numpy import ndarray, int64, float64, int32, float32

    _numpy_ = True
except ModuleNotFoundError:
    _numpy_ = False
try:
    from pandas import Series, DataFrame

    _pandas_ = True
except ModuleNotFoundError:
    _pandas_ = False

if _numpy_ and _pandas_:
    number_ = (int, float, complex, int32, int64, float32, float64)
    number = Union[int, float, complex, int32, int64, float32, float64]
    numeric_ = (int, float, complex, int32, int64, float32, float64, ndarray, Series, DataFrame)
    numeric = Union[int, float, complex, int32, int64, float32, float64, ndarray, Series, DataFrame]
    array_like_ = (ndarray, Series, DataFrame)
    array_like = Union[ndarray, Series, DataFrame]
elif _numpy_:
    number_ = (int, float, complex, int32, int64, float32, float64)
    number = Union[int, float, complex, int32, int64, float32, float64]
    numeric_ = (int, float, complex, ndarray, int32, int64, float32, float64)
    numeric = Union[int, float, complex, ndarray, int32, int64, float32, float64]
    array_like_ = (ndarray, )
    array_like = ndarray
elif _pandas_:
    number_ = (int, float, complex)
    number = Union[int, float, complex]
    numeric_ = (int, float, complex, Series, DataFrame)
    numeric = Union[int, float, complex, Series, DataFrame]
    array_like_ = (Series, DataFrame)
    array_like = Union[Series, DataFrame]
else:
    number_ = (int, float, complex)
    number = Union[int, float, complex]
    numeric_ = (int, float, complex)
    numeric = Union[int, float, complex]
    array_like_ = (None, )
    array_like = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:38:58 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.4'
__release__ = 20240521
__all__ = ['unit_product', 'unit_division', 'unit_base_power', 'unit_power', 'unit_addition', 'unit_inverse']

from .dictionaries import dictionary, unitless_names
from .converter import convertible
from .helpers.is_number import is_number
from .helpers.unit_string_tools import reduce_units
from .helpers.multi_split import multi_split


_super_numbers_ = dict(zip('⁰¹³²⁴⁵⁶⁷⁸⁹', '0123456789'))

def unit_split(unit_string: str) -> str:
    us = multi_split(unit_string)
    return us


def unit_base_power(unit_string: str) -> tuple:  # tuple[str, int]
    u_bas, u_pow = '', ''
    oth = ''
    inv = False
    if '/' in unit_string and len(unit_string.split('/')) == 2 and is_number(unit_string.split('/')[0]):
        inv_pow, u_bas = unit_string.split('/')
        inv = True
    else:
        for c in unit_string:
            if c.isdigit():
                u_pow += oth + c
                oth = ''
            elif c in ['-', '+', '.']:
                oth += c
            else:
                if inv and len(u_pow) > 0:
                    if u_pow[0] == '-':
                        u_pow = u_pow[1:]
                    else:
                        u_pow = '-' + u_pow
                if c in ['/']:
                    inv = not inv
                u_bas += oth + c
                oth = ''
        if inv and len(u_pow) > 0:
            if u_pow[0] == '-':
                u_pow = u_pow[1:]
            else:
                u_pow = '-' + u_pow
        inv = False
    u_pow = 1 if u_pow == '' else float(u_pow) if '.' in u_pow else int(u_pow)
    if inv:
        u_pow = u_pow * -1 * (float(inv_pow) if '.' in inv_pow else int(inv_pow))
    return u_bas, u_pow


def unit_base(unit_string: str) -> str:
    return unit_base_power(unit_string)[0]


def unit_product(unit_string1: str, unit_string2: str) -> str:
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == \
            unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == \
            unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        if unit_string2.lower().strip(' ()') not in unitless_names:  # dictionary['Dimensionless']:
            return unit_string2
        else:
            return unit_string1

    if unit_string1 == unit_string2:
        parenthesis = False
        for s in '*/+-':
            if s in unit_base_power(unit_string1)[0]:
                parenthesis = True
                break
        if parenthesis:
            return '(' + unit_string1 + ')2'
        else:
            return unit_base_power(unit_string1)[0] + str(int(unit_base_power(unit_string1)[1]) * 2)

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unit_product(unit_string1, unit_string1)

    u1bas, u1pow = unit_base_power(unit_string1)
    u2bas, u2pow = unit_base_power(unit_string2)

    if u1pow == -1 and u2pow != -1:
        return unit_division(unit_string2, u1bas)
    elif u1pow != -1 and u2pow == -1:
        return unit_division(unit_string1, u2bas)

    if convertible(u1bas, u2bas):
        u_pow = u1pow + u2pow
        if u_pow == -1:
            result = u1bas + '-1'
        elif u_pow == 1:
            result = u1bas
        elif u_pow == 0:
            result = u1bas + '/' + u1bas
        else:
            for c in ['+', '-', '^']:  # '*','/'
                if c in u1bas:
                    u1bas = '(' + u1bas + ')'
                    break
            result = u1bas + str(u_pow)
    else:
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '(' + unit_string1 + ')'
                break
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '(' + unit_string2 + ')'
                break
        result = unit_string1 + '*' + unit_string2
    return reduce_units(result, raise_error=False)


def unit_division(unit_string1: str, unit_string2: str) -> str:
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == \
            unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == \
            unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        if unit_string2.lower().strip(' ()') not in unitless_names:  # dictionary['Dimensionless']:
            u_bas, u_pow = unit_base_power('1/' + unit_string2)
            return u_bas + str(u_pow)
        else:
            return unit_string1

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unit_division(unit_string1, unit_string1)

    u1bas, u1pow = unit_base_power(unit_string1)
    u2bas, u2pow = unit_base_power(unit_string2)

    if u1pow == -1 and u2pow != -1:
        return unit_product(u1bas, unit_string2)
    elif u1pow != -1 and u2pow == -1:
        return unit_product(unit_string1, u2bas)

    if convertible(u1bas, u2bas):
        u_pow = u1pow - u2pow
        if u_pow == -1:
            result = u1bas + '-1'
        elif u_pow == 1:
            result = u1bas
        elif u_pow == 0:
            result = u1bas + '/' + u1bas
        else:
            for c in ['+', '-', '^']:  # '*','/'
                if c in u1bas:
                    u1bas = '(' + u1bas + ')'
                    break
            result = u1bas + str(u_pow)

    elif ('+' not in unit_string1 and '-' not in unit_string1 and '^' not in unit_string1) and (
            '+' not in unit_string2 and '-' not in unit_string2 and '^' not in unit_string2) and (
            '*' in unit_string1 and unit_base(unit_string2) in map(unit_base, unit_string1.split('*'))):
        result = ''
        for u in unit_string1.split('*'):
            if unit_string2 == u:
                return '*'.join([u for u in unit_string1.split('*') if u != unit_string2])
            elif unit_base(unit_string2) == unit_base(u):
                result = (result + '*' + unit_division(u, unit_string2).strip('*')).strip('*')
            else:
                result = (result + '*' + u).strip('*')
    else:
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '(' + unit_string1 + ')'
                break
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '(' + unit_string2 + ')'
                break
        result = unit_string1 + '/' + unit_string2
    return reduce_units(result, raise_error=False)


def unit_power(unit_string: str, power: int or str) -> str:
    if unit_string is None:
        unit_string = 'dimensionless'
    if power is None:
        power = 1

    if type(unit_string) is str and len(unit_string.strip(' ()')) == 0:
        unit_string = 'dimensionless'
    if type(power) is str and len(power.strip(' ()')) == 0:
        power = 1

    u1bas, u1pow = unit_base_power(unit_string)

    if type(power) in (int, float):
        u_pow = u1pow * power
        if u_pow == 0:
            return 'dimensionless'
        elif u_pow == 1:
            return u1bas
        else:
            return u1bas + str(u_pow)
    elif type(power) is str:
        u2bas, u2pow = unit_base_power(power)
        u_pow = u1pow * u2pow
        if u_pow == 0:
            return 'dimensionless'
        elif u_pow == 1:
            return u1bas + '^' + u2bas
        else:
            return u1bas + '^' + str(u_pow) + u2bas
    else:
        raise TypeError('Power must be units string or numeric.')


def unit_addition(unit_string1:str, unit_string2:str) -> str:
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        if unit_string2.lower().strip(' ()') not in unitless_names:  # dictionary['Dimensionless']:
            return unit_string2
        else:
            return unit_string1

    if convertible(unit_string2, unit_string1):
        return unit_string1
    elif convertible(unit_string1, unit_string2):
        return unit_string2
    else:
        return unit_string1 + '+' + unit_string2


def unit_inverse(unit_string: str) -> str:
    if unit_string is None:
        unit_string = 'dimensionless'

    if type(unit_string) is str and len(unit_string.strip(' ()')) == 0:
        unit_string = 'dimensionless'

    if unit_string.lower().strip(' ()') in unitless_names:  # dictionary['Dimensionless']:
        return unit_string

    ubas, upow = unit_base_power(unit_string)
    if upow == -1:
        return ubas
    elif upow < -1:
        return ubas + str(abs(upow))
    elif upow > 1:
        return ubas + str(-1 * upow)

    if '/' not in unit_string:
        return ubas + '-1'  # '1/' + ubas
    elif len(unit_string.split('/')) == 2:
        return unit_string.split('/')[1] + '/' + unit_string.split('/')[0]
    elif len(unit_string.split('/')) == 3:
        return unit_string.split('/')[2] + '*' + unit_string.split('/')[1] + '/' + unit_string.split('/')[0]
    else:  # wtf?
        return '/'.join(unit_string.split('/')[::-1])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['Length', 'Area', 'Volume', 'Permeability']

from ..dictionaries import dictionary as _dictionary
from ..unit_class import Unit
from ..helpers.common_classes import unit_or_str, numeric


class Length(Unit):
    class_units = _dictionary['Length']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'length' if name is None else name
        self.kind = Length
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Area(Unit):
    class_units = _dictionary['Area']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'area' if name is None else name
        self.kind = Area
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Volume(Unit):
    class_units = _dictionary['Volume']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'volume' if name is None else name
        self.kind = Volume
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Permeability(Unit):
    class_units = _dictionary['Permeability']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'permeability' if name is None else name
        self.kind = Permeability
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



# -*- coding: utf-8 -*-
"""
Created on Sun May 23 09:56:26 2021

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

from unyts.operations import unit_base_power, unit_product, unit_division, unit_power


def test_unit_base_power():
    assert unit_base_power('m') == ('m', 1)
    assert unit_base_power('ft') == ('ft', 1)
    assert unit_base_power('m2') == ('m', 2)
    assert unit_base_power('ft3') == ('ft', 3)
    assert unit_base_power('1/m') == ('m', -1)
    assert unit_base_power('1/ft') == ('ft', -1)
    assert unit_base_power('day-1') == ('day', -1)


def test_unit_product():
    assert unit_product('m', 'ft') == 'm2'
    assert unit_product('ft', 'm') == 'ft2'
    assert unit_product('m', 's') == 'm*s'
    assert unit_product('in2', 'in') == 'in3'


def test_unit_division():
    assert unit_division('m', 'm') == 'm/m'
    assert unit_division('m', 'm2') == 'm-1'
    assert unit_division('cm3', 'cm') == 'cm2'


def test_unit_power():
    assert unit_power('m', -1) == 'm-1'
    assert unit_power('m', 0) == 'dimensionless'
    assert unit_power('m', 1) == 'm'
    assert unit_power('m', 2) == 'm2'
    assert unit_power('m', 3) == 'm3'
    assert unit_power('ft2', -1) == 'ft-2'
    assert unit_power('ft2', 0) == 'dimensionless'
    assert unit_power('ft2', 1) == 'ft2'
    assert unit_power('ft2', 2) == 'ft4'
    assert unit_power('ft2', 3) == 'ft6'
    assert unit_power('s', 's') == 's^s'
    assert unit_power('m', 's') == 'm^s'
    assert unit_power('m2', 'm') == 'm^2m'
    assert unit_power('m2', 'm2') == 'm^4m'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['Rate', 'Speed', 'Velocity', 'Acceleration']

from ..dictionaries import dictionary as _dictionary
from ..unit_class import Unit
from ..helpers.common_classes import unit_or_str, numeric


class Rate(Unit):
    class_units = _dictionary['Rate']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'rate' if name is None else name
        self.kind = Rate
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Velocity(Unit):
    class_units = _dictionary['Velocity']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'velocity' if name is None else name
        self.kind = Velocity
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


def Speed(value: numeric, units: unit_or_str, name=None):
    return Velocity(value, units, name)


class Acceleration(Unit):
    class_units = _dictionary['Acceleration']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'acceleration' if name is None else name
        self.kind = Acceleration
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 12:36:48 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""
import logging
from os.path import isfile

from .errors import NoFVFError
from .parameters import unyts_parameters_

try:
    from cloudpickle import dump as cloudpickle_dump, load as cloudpickle_load

    _cloudpickle_ = True
except ModuleNotFoundError:
    _cloudpickle_ = False

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

__version__ = '0.4.18'
__release__ = 20240823
__all__ = ['UNode', 'UDigraph', 'Conversion']


class UNode(object):
    __slots__ = ('name',)

    def __init__(self, name):
        self.name = name if type(name) is str else ''

    def get_name(self):
        return self.name

    def __str__(self):
        return self.name


class UDigraph(object):
    """
    A class to store the digraph containing the units network and methods to facilitate its use.
    The `edges` attribute is a dict mapping each node to a list of its children
    The `children_of` method returns the list of nodes with direct relation to the key node. This method will be used by the search algorithms.
    """
    __slots__ = ('edges', '_edges_str', 'previous', 'recursion_limit', 'fvf', 'memory', 'print', '_cloudpickle_')

    def __init__(self) -> None:
        self.edges = {}
        self._edges_str = None
        self.previous = [(None, None)]
        self.recursion_limit = 5
        self.fvf = None
        self.memory = {}
        self.print = False
        self._cloudpickle_ = _cloudpickle_
        if unyts_parameters_.cache_ and unyts_parameters_.memory_:
            self.load_memory()

    def get_edges_str(self) -> dict:
        if self._edges_str is None:
            self._edges_str = {str(k): {str(each) for each in v[0]} for k, v in self.edges.items()}
        return self._edges_str

    def save_memory(self, path=None) -> None:
        if path is None:
            path = unyts_parameters_.get_user_folder() + 'search_memory.cache'
        if self._cloudpickle_:
            logging.info('saving search memory to cache...')
            with open(path, 'wb') as f:
                cloudpickle_dump(self.memory, f)
        else:
            logging.warning("Missing `cloudpickle` package. Not able to cache search memory.")

    def load_memory(self, path=None) -> None:
        if path is None:
            path = unyts_parameters_.get_user_folder() + 'search_memory.cache'
        if not self._cloudpickle_:
            logging.warning("Missing `cloudpickle` package. Not able to cache search memory.")
        if not isfile(unyts_parameters_.get_user_folder() + 'search_memory.cache'):
            msg = "starting clean memory..."
            logging.info(msg)
        else:
            logging.info('loading memory from cache...')
            try:
                with open(path, 'rb') as f:
                    cached_memory = cloudpickle_load(f)
                    self.memory.update(cached_memory)
                    msg = f"{len(self.memory)} conversion path{'' if len(self.memory) == 1 else 's'} in memory."
                    logging.info(msg)
            except:
                msg = 'Failed to load memory from cache.'
                logging.error(msg)
        unyts_parameters_.last_path_str = msg

    def clean_memory(self):
        self.memory = {}

    def add_node(self, node) -> None:
        if node in self.edges:
            raise ValueError('Duplicate node')
        else:
            self.edges[node] = [], []

    def add_edge(self, edge, reverse=False) -> None:
        src = edge.get_source()
        dest = edge.get_destination()
        conv = edge.get_convert()
        if not (src in self.edges and dest in self.edges):
            raise ValueError('Node not in graph')
        if dest not in self.edges[src][0]:  # avoid duplication
            self.edges[src][0].append(dest)
            self.edges[src][1].append(conv)

    def children_of(self, node):
        return self.edges[node][0]

    def has_node(self, node):
        if type(node) is str:
            return node in [n.get_name() for n in self.edges]
        else:
            return node in self.edges

    def get_node(self, name):
        result = [n for n in self.edges if n.get_name() == name]
        if len(result) == 0:
            raise NameError(name)
        else:
            return result[0]

    def list_nodes(self):
        return list(set([n.get_name() for n in self.edges.keys()]))

    def convert(self, value, src, dest):
        if type(src) != UNode:
            src = self.get_node(src)
        if type(dest) != UNode:
            dest = self.get_node(dest)
        return self.edges[src][1][self.edges[src][0].index(dest)](value)

    def conversion(self, src, dest):
        if type(src) != UNode:
            src = self.get_node(src)
        if type(dest) != UNode:
            dest = self.get_node(dest)
        return self.edges[src][1][self.edges[src][0].index(dest)]

    def __str__(self) -> str:
        result = ''
        for src in self.edges:
            for dest in self.edges[src]:
                result = result + src.get_name() + '->' \
                         + dest.get_name() + \
                         str(self.conv) + '\n'
        return result[:-1]  # remove final \n

    def set_fvf(self, FVF) -> None:
        if type(FVF) is str:
            try:
                FVF = float(FVF)
            except ValueError:
                print(f'received FVF value is not a number: {str(FVF)}')
        if type(FVF) in (int, float):
            if FVF <= 0:
                logging.error('FVF should be a positive number...')
            self.fvf = FVF

    def get_fvf(self):
        def valid_fvf(FVF):
            if type(FVF) is str:
                try:
                    FVF = float(FVF)
                except ValueError:
                    return False
            if type(FVF) in (int, float):
                if FVF <= 0:
                    return False
                else:
                    return FVF
            else:
                return False

        if self.fvf is None:
            if unyts_parameters_.gui:
                raise NoFVFError()
            else:
                print('Please enter formation Volume factor (FVF) in reservoir_volume/standard_volume:')
                while self.fvf is None:
                    self.fvf = input(' FVF (rV/stV) = ')
                    if not valid_fvf(self.fvf):
                        self.fvf = None
                    else:
                        self.fvf = valid_fvf(self.fvf)
        return self.fvf


class Conversion(object):
    __slots__ = ('src', 'dest', 'conv', 'rev', 'alias')

    def __init__(self, src, dest, conv, reverse=False, alias=False):
        """Assumes src and dest are nodes"""
        self.src = src
        self.dest = dest
        self.conv = conv
        self.rev = reverse
        self.alias = alias

    def get_source(self):
        return self.src

    def get_destination(self):
        return self.dest

    def convert(self, value):
        return self.conv(value)

    def reverse(self, value):
        return value / self.conv(1)

    def get_convert(self):
        if self.rev and self.conv is not None:
            return lambda x: x / self.conv(1)
        else:
            return self.conv

    def __str__(self) -> str:
        return self.src.get_name() + '->' + self.dest.get_name()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 22 00:23:43 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['Data']

from ..dictionaries import dictionary as _dictionary
from ..unit_class import Unit
from ..helpers.common_classes import unit_or_str, numeric


class Data(Unit):
    class_units = _dictionary['Data']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'data' if name is None else name
        self.kind = Data
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 21:34:21 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

from unyts import convert
from pandas import read_excel
from math import isnan
from unyts.converter import _apply_conversion, _get_conversion, _converter, _clean_print_conversion_path, convert_for_SimPandas, convertible
from unyts.database import units_network as UnNe
import numpy as np


def test__get_conversion():
    assert _get_conversion(3, 'meter', 'meter') == (3, [UnNe.get_node('meter')])
    assert _get_conversion('16/06/1969', 'date', 'DATE') == ('16/06/1969', [UnNe.get_node('date'), UnNe.get_node('DATE')])
    assert _get_conversion(2, 'foot', None) == (2, [])
    assert _get_conversion(3, None, 'meter') == (3, [])
    assert _get_conversion(4, None, None) == (4, [None])
    assert _get_conversion(5, 'fraction', 'meter') == (5, [])
    assert _get_conversion(0.25, 'fraction', 'percent') == (25.0, ['*', 100])
    assert _get_conversion(0.25, 'percent', 'fraction') == (0.0025, ['/', 100])
    assert _get_conversion(0.25, 'dimensionless', 'fraction') == (0.25, [])
    assert _get_conversion(0.01, 'F/ft', 'C/m') == (0.018226888305628464, ['*', 0.5555555555555556, '/', 1,
                                                                           UnNe.get_node('ft'),
                                                                           UnNe.get_node('foot'),
                                                                           UnNe.get_node('yard'),
                                                                           UnNe.get_node('meter'),
                                                                           UnNe.get_node('m')])
    assert _get_conversion(0.33, 'fraction', 'ft/ft') == (0.33, [])
    assert _get_conversion(0.33, 'ft3/ft3', 'fraction') == (0.33, [])
    assert _get_conversion(10, 'scf/stb', 'stb/scf') == (1/10, ['1/'])
    assert _get_conversion(1, 'm', 'cm') == (100, [UnNe.get_node('m'), UnNe.get_node('cm')])
    assert _get_conversion(1, 'meter', 'inch') == (39.37007874015748,
                                                   [UnNe.get_node('meter'),
                                                    UnNe.get_node('yard'),
                                                    UnNe.get_node('foot'),
                                                    UnNe.get_node('inch')])


def test__converter():
    assert _converter(0.433, 'psi/ft', 'bar/m') == (0.09794717545606699, [UnNe.get_node('psi'),
                                                                          UnNe.get_node('bar'),
                                                                          '/', 1,
                                                                          UnNe.get_node('ft'),
                                                                          UnNe.get_node('foot'),
                                                                          UnNe.get_node('yard'),
                                                                          UnNe.get_node('meter'),
                                                                          UnNe.get_node('m')])


def test__apply_conversion():
    conv, conv_path = _get_conversion(3, 'meter', 'meter')
    assert _apply_conversion(3, conv_path) == conv

    conv, conv_path = _get_conversion(0.25, 'fraction', 'percent')
    assert _apply_conversion(0.25, conv_path) == conv

    conv, conv_path = _get_conversion(0.25, 'percent', 'fraction')
    assert _apply_conversion(0.25, conv_path) == conv

    conv, conv_path = _get_conversion(0.25, 'dimensionless', 'fraction')
    assert _apply_conversion(0.25, conv_path) == conv

    conv, conv_path = _get_conversion(10, 'scf/stb', 'stb/scf')
    assert _apply_conversion(10, conv_path) == conv

    conv, conv_path = _get_conversion(0.01, 'F/ft', 'C/m')
    assert _apply_conversion(0.01, conv_path) == conv

    conv, conv_path = _get_conversion(1, 'meter', 'inch')
    assert _apply_conversion(1, conv_path) == conv

    conv, conv_path = _get_conversion(0.433, 'psi/ft', 'bar/m')
    assert round(_apply_conversion(0.433, conv_path), 6) == round(conv, 6)


def test__clean_print_conversion_path():
    assert _clean_print_conversion_path() in (True, False)
    assert _clean_print_conversion_path(True) is True
    assert _clean_print_conversion_path(False) is False
    assert _clean_print_conversion_path(1) is True
    assert _clean_print_conversion_path(0) is False


def test_convert():
    assert convert(1, 'unitless', np.nan) == 1
    assert convert(1, 'unitless', None) == 1
    assert convert(1, None, np.nan) == 1
    assert convert(1, np.nan, np.nan) == 1
    assert convert(1, np.nan, 'unitless') == 1

    data = read_excel('./tests/conversions_check.xlsx')
    error = 1E-4
    for i in data.index:
        if not isnan(data.loc[i, 'out']):
            print(data.loc[i, 'id'], ': converting', data.loc[i, 'source'], 'into', data.loc[i, 'target'])
            result = convert(data.loc[i, 'in'], data.loc[i, 'source'], data.loc[i, 'target'], False)
            print(result, data.loc[i, 'out'])
            assert abs(1 - result / data.loc[i, 'out']) < error


def test_convert_for_SimPandas():
    array = np.random.rand(10)
    assert (convert_for_SimPandas(array, 'meter', 'litre') == array).all()
    assert (convert_for_SimPandas(array, 'm', 'yd') == convert(array, 'm', 'yd')).all()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['Density', 'VolumeRatio', 'ProductivityIndex', 'PressureGradient']

from ..dictionaries import dictionary as _dictionary
from ..unit_class import Unit
from ..helpers.common_classes import unit_or_str, numeric


class Density(Unit):
    class_units = _dictionary['Density']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'density' if name is None else name
        self.kind = Density
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class VolumeRatio(Unit):
    class_units = _dictionary['VolumeRatio']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'volume_ratio' if name is None else name
        self.kind = VolumeRatio
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class ProductivityIndex(Unit):
    class_units = _dictionary['ProductivityIndex']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'productivity_index' if name is None else name
        self.kind = ProductivityIndex
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class PressureGradient(Unit):
    class_units = _dictionary['PressureGradient']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, units: unit_or_str, name=None):
        self.name = 'pressure_gradient' if name is None else name
        self.kind = PressureGradient
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 17:52:34 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.6.1'
__release__ = 20240612
__all__ = ['BFS', 'lean_BFS', 'DFS', 'print_path']

import logging
from unyts import unyts_parameters_


def BFS(graph, start, end, verbose=False) -> list:
    """
    Implementation of Breadth-First Search algorithm.
    Assumes graph is a digraph; `start` and `end` are nodes in the graph network.
    Returns a shortest path from `start` to `end` in graph.

    Parameters
    ----------
    graph: Digraph
    start: node
    end: node
    verbose: bool
        to print or not print messages.
    Returns
    -------
    shortest_path: list
    """
    init_path = [start]
    path_queue = [init_path]
    visited = list()
    while len(path_queue) != 0:
        # get and remove oldest element in path_queue
        conv_path = path_queue.pop(0)
        if conv_path in visited:
            if verbose:
                logging.info(f"""<BFS>: {len(path_queue)} paths in queue. Already visited BFS path:\n{print_path(conv_path)}""")
        else:
            if verbose:
                logging.info(f"""<BFS> {len(path_queue)} paths in queue. Current BFS path:\n{print_path(conv_path)}""")
            last_node = conv_path[-1]
            if last_node is end:
                if verbose:
                    logging.info(f"""<BFS> Found end node {end.get_name()} in the path:\n{print_path(conv_path)}""")
                return conv_path
            path_queue += [conv_path + [next_node]
                           for next_node in graph.children_of(last_node) 
                           if next_node not in conv_path]
            visited.append(conv_path)


def DFS(graph, start, end, verbose=False, branch_depht=25) -> list:
    """
    Implementation of Depth-First Search algorithm.
    Assumes graph is a digraph; `start` and `end` are nodes in the graph network.
    Returns a first found path from `start` to `end` in graph.

    Parameters
    ----------
    graph: Digraph
    start: node
    end: node
    verbose: bool
        to print or not print messages.
    Returns
    -------
    shortest_path: list
    """
    from unyts.converter import _get_descendants
    branch_depht = unyts_parameters_.generations_limit() if branch_depht is None else branch_depht
    def dfs_(graph, node, visited, path_queue):
        visited.add(node)
        for child in graph.children_of(node):
            this_path = []
            if child in visited:
                continue
            if end.get_name() not in _get_descendants(child.get_name(), branch_depht):
                continue
            else:
                this_path.append(child)
                if child is end:
                    return path_queue + this_path
                else:
                    this_path = dfs_(graph, child, visited, this_path)
                    if end in this_path:
                        path_queue += this_path
                        break
        return path_queue
    path_queue = [start]
    visited = set()
    path_queue = dfs_(graph, start, visited, path_queue)
    if end in path_queue:
        return path_queue


class SlimUDigraph(object):
    """
    A simplified class, from the UDgraph class, to store a slimmed digraph dictionary containing only units related to the start and end of the search.
    The `edges` attribute is a dict mapping each node to a list of its children.
    The `children_of` method returns the list of nodes with direct relation to the key node. This method will be used by the search algorithms.
    """
    def __init__(self, edges:dict={}):
        self.edges = edges

    def children_of(self, node):
        return self.edges[node][0] if node in self.edges else []


def lean_BFS(graph, start, end, verbose=False, max_generations_screening=25) -> list:
    """
    Runs BFS algorithm on a lean digraph network, where only the nodes related to the `start` and `end` nodes are kept.
    Assumes graph is a digraph; `start` and `end` are nodes in the graph network.
    Returns a shortest path from `start` to `end` in graph.

    Parameters
    ----------
    graph: Digraph
    start: node
    end: node
    verbose: bool
        to print or not print messages.
    Returns
    -------
    shortest_path: list
    """
    from unyts.converter import _get_descendants
    max_generations_screening = unyts_parameters_.generations_limit() if max_generations_screening is None else max_generations_screening
    generations_list = [g for g in [0, 1, 2, 3, 4, 5, 7, 10, 13, 16, 20, 25, 30, 40, 50] if g < max_generations_screening] + [max_generations_screening]
    selection = set()
    for generations in generations_list:
        generations += 1
        start_descendants = _get_descendants(start.get_name(), generations)
        end_descendants = _get_descendants(end.get_name(), generations)
        selection = start_descendants.intersection(end_descendants)
        if len(selection) > 0:
            break
    selection = selection.union({node for each in selection for node in _get_descendants(each, generations, get_combinations=False)})
    selected_edges = {k: v for k, v in graph.edges.items() if k.get_name() in selection}
    if len(selected_edges) > 0:
        if verbose:
            logging.info(f"<lean BFS> search graph slimmed from {len(graph.edges)} to {len(selected_edges)} nodes, in {generations} generations.")
    else:
        return None
    slim_graph = SlimUDigraph(selected_edges)
    return BFS(slim_graph, start, end, verbose=verbose and unyts_parameters_.verbose_details_ > 0)


def print_path(path: list) -> str:
    """
    Assumes path is a list of nodes.

    Parameters
    ----------
    path: list
        a list of nodes
    Returns
    -------
    str
        string representation of list of nodes
    """
    result = '    '
    if len(path) == 1:
        result = f"{path[0]} = {path[0]}"
    else:
        for i in range(len(path)):
            if type(path[i]) is str and path[i] not in ['(1)', '(v)']:
                if result[-3:] == ' > ':
                    result = result[:-3]
                result = result + ' ' + path[i] + ' '
            elif type(path[i]) in (int, float, complex):
                result = result + str(path[i]) + ' '
            else:
                result = result + str(path[i])
                if i != len(path) - 1:
                    result = result + ' > '
    return result.strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 15:57:27 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.8.3'
__release__ = 20240612
__all__ = ['convert', 'convertible']

from .database import units_network
from .dictionaries import dictionary, temperatureRatioConversions, uncertain_names
from .Empty import Empty, str_Empty
from .searches import BFS, lean_BFS, DFS, print_path
from .errors import NoConversionFoundError
from .parameters import unyts_parameters_, _get_density
from .helpers.unit_string_tools import split_unit as _split_unit, reduce_parentheses as _reduce_parentheses
from functools import reduce
from typing import Union
from sys import getrecursionlimit
import logging

try:
    import numpy as np
    from numpy import ndarray

    _numpy_ = True
except ModuleNotFoundError:
    _numpy_ = False
    logging.warning("Missing NumPy package, operations with `list` of values will fail.")
try:
    from pandas import Series, DataFrame

    _pandas_ = True
except ModuleNotFoundError:
    _pandas_ = False

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

if _numpy_ and _pandas_:
    _numeric = (int, float, complex, ndarray, Series, DataFrame)
    numeric = Union[int, float, complex, ndarray, Series, DataFrame]
elif _numpy_:
    _numeric = (int, float, complex, ndarray)
    numeric = Union[int, float, complex, ndarray]
elif _pandas_:
    _numeric = (int, float, complex, Series, DataFrame)
    numeric = Union[int, float, complex, Series, DataFrame]
else:
    _numeric = (int, float, complex)
    numeric = Union[int, float, complex]


def _str2lambda(string: str):
    """
    Returns a lambda with a multiplication or division operation, depending on the received string.

    Parameters
    ----------
    string: str

    Returns
    -------
    lambda
    """
    if string == '/':
        return lambda x, y: x / y
    if string == '*':
        return lambda x, y: x * y


def _apply_conversion(value, conversion_path):
    """
    Helper function to apply the functions stated in the `conversion_path`.
    Parameters
    ----------
    value: numeric (int, float, NumPy.array, ...)
        the value(s) to apply the conversion
    conversion_path: list
        conversion_path provided by _get_conversion()

    Returns
    -------
        numeric (int, float, NumPy.array, ...)
    """
    if len(conversion_path) == 1 and conversion_path[0] == '1/':
        return 1 / value
    i = 0
    while i < len(conversion_path) - 1:
        this_step, next_step = conversion_path[i], conversion_path[i + 1]
        if type(this_step) is str:
            if len(this_step) == 1:
                if type(next_step) in (int, float, complex):
                    value = _str2lambda(this_step)(value, _apply_conversion(next_step, conversion_path[i + 2:]))
                else:
                    value = _str2lambda(this_step)(value, _apply_conversion(value, conversion_path[i + 1:]))
                break
            elif this_step == '1/':
                value = 1 / _apply_conversion(value, conversion_path[i + 1:])
                break
            else:
                raise ValueError("string operation in conversion_path must be '/' or '*'")
        elif type(this_step) in (int, float, complex):
            if type(next_step) is str:
                value = _str2lambda(next_step)(value, _apply_conversion(this_step, conversion_path[i + 2:]))
                break
            else:
                value = _apply_conversion(this_step, conversion_path[i + 1:])
                break
        elif type(next_step) is str:
            if len(next_step) == 1:
                if len(conversion_path) == i + 2:
                    further_step = conversion_path[i + 2]
                    if type(further_step) in (int, float, complex):
                        value = _str2lambda(next_step)(value, further_step)
                    else:
                        value = _str2lambda(next_step)(value, _apply_conversion(value, further_step))
                    break
                elif len(conversion_path) > i + 2:
                    further_step = conversion_path[i + 2]
                    if type(further_step) in (int, float, complex):
                        value = _str2lambda(next_step)(value, _apply_conversion(further_step, conversion_path[i + 3:]))
                    else:
                        value = _str2lambda(next_step)(value, _apply_conversion(value, further_step))
                    break
        else:
            value = units_network.convert(value, this_step, next_step)
        i += 1
    return value


def _lambda_conversion(conversion_path):
    """
    Helper function to make a lambda applying in sequence all the conversions listed in the conversion_path.

    Parameters
    ----------
    conversion_path: list
        sequence of the conversions to be applied
    Returns
    -------
    conversion_lambda: lambda
        to apply the conversion path
    """
    big_lambda = [units_network.conversion(conversion_path[i], conversion_path[i + 1])
                  for i in range(len(conversion_path) - 1)]
    return lambda x: _lambda_loop(x, big_lambda[:])


def _lambda_loop(x, lambda_list):
    """
    Helper function to apply lambda functions in sequence.

    Parameters
    ----------
    x: numeric
        the value to apply the conversion sequence
    lambda_list : list
        a list of the conversions that should be applied in sequence

    Returns
    -------
    lambda_sequence : lambda
        all the lambdas in the list applied in sequence
    """
    for lambda_i in lambda_list:
        x = lambda_i(x)
    return x


def _get_pair_child(unit: str):
    """
    Returns the following child of a unit node.
    Parameters
    ----------
    unit: str

    Returns
    -------
        unit_node
    """
    # get the Unit node if the name received is string
    unit = units_network.get_node(unit) if type(unit) is str else unit

    # get a pair of units children
    pair_child = list(filter(lambda u: '/' in u or '*' in u, [u.get_name() for u in units_network.children_of(unit)]))

    # if a pair of units child is found, return the one with the shorter name
    if len(pair_child) > 0:
        pair_child = sorted(pair_child, key=len)[0]
    # if no children found at this level, look for children in next level
    else:
        for child in units_network.children_of(unit):
            pair_grandchild = list(
                filter(lambda u: '/' in u or '*' in u, [u.get_name() for u in units_network.children_of(child)]))
            if len(pair_grandchild) > 0:
                pair_child = sorted(pair_grandchild, key=len)[0]
                break

    if type(pair_child) is str:
        return pair_child


def _get_recursion_limit(recursion=None):
    if recursion is None:
        recursion = min(getrecursionlimit() - 15, unyts_parameters_.max_recursion_)
    elif recursion > 1:
        recursion = min(getrecursionlimit() - 15, recursion) - 1
    elif recursion <= 0:
        recursion = -1
    else:
        recursion -= 1
    return recursion


def _get_conversion(value, from_unit, to_unit, recursion=None):
    """
    Helper function to handle looking for the conversion factor of special cases and through the units network.

    Parameters
    ----------
    value: numeric or None
    from_unit: str
    to_unit: str

    Returns
    -------
        (conversion, conversion_path)

    """
    # get and set recursion limit
    recursion = _get_recursion_limit(recursion)
    if unyts_parameters_.verbose_ and unyts_parameters_.verbose_details_ >= 2:
        logging.info(f"_get_conversion: {recursion} remaining recursions, converting from {from_unit} to {to_unit}")
    if recursion < 0:
        return None, None


    # check if already solved and memorized
    if (from_unit, to_unit) in units_network.memory:
        conversion_lambda, conversion_path = units_network.memory[(from_unit, to_unit)]
        return (conversion_lambda, conversion_path) if (conversion_lambda is None or value is None) \
            else (conversion_lambda(value), conversion_path)

    ## specific cases for quick conversions ##
    # no conversion required if 'from' and 'to' units are the same units
    if from_unit == to_unit:
        if value is None:
            return lambda x: x, [units_network.get_node(from_unit) if units_network.has_node(from_unit) else from_unit]
        else:
            return value, [units_network.get_node(from_unit) if units_network.has_node(from_unit) else from_unit]

    # no conversion required if 'from' and 'to' units are dates
    if from_unit in dictionary['Date'] and to_unit in dictionary['Date']:
        if value is None:
            return lambda x: x, [units_network.get_node(from_unit) if units_network.has_node(from_unit) else from_unit,
                                 units_network.get_node(to_unit) if units_network.has_node(to_unit) else to_unit]
        else:
            return value, [units_network.get_node(from_unit) if units_network.has_node(from_unit) else from_unit,
                           units_network.get_node(to_unit) if units_network.has_node(to_unit) else to_unit]

    # from None to some units or vice-versa
    if from_unit is None or to_unit is None:
        return (lambda x: x, []) if value is None else (value, [])

    # from Dimensionless to Percentage or vice-versa
    if (from_unit.lower() in dictionary['Dimensionless']) and to_unit.lower() in dictionary['Percentage']:
        return (lambda x: x * 100, ['*', 100]) if value is None else (value * 100, ['*', 100])
    if from_unit.lower() in dictionary['Percentage'] and to_unit.lower() in dictionary['Dimensionless']:
        return (lambda x: x / 100, ['/', 100]) if value is None else (value / 100, ['/', 100])

    # from Dimensionless to some units (not ratios), to allow assign units to Dimensionless numbers
    if from_unit.lower() in dictionary['Dimensionless'] and '/' not in to_unit:
        return (lambda x: x, []) if value is None else (value, [])

    # special case for Temperature ratios
    if '/' in from_unit and len(from_unit.split('/')) == 2 and from_unit.split('/')[0] in dictionary['Temperature'] \
            and '/' in to_unit and len(to_unit.split('/')) == 2 and to_unit.split('/')[0] in dictionary['Temperature']:
        t1, d1 = from_unit.split('/')
        t2, d2 = to_unit.split('/')
        num = temperatureRatioConversions[(t1, t2)]
        den, den_path = _get_conversion(1, d1, d2, recursion=recursion)
        if num is None or den is None:
            if unyts_parameters_.raise_error_:
                raise NoConversionFoundError("from '" + str(d1) + "' to '" + str(d2) + "'")
            else:
                return None, None
        den_path = [1] + den_path
        if value is None:
            return lambda x: x * num / den, ['*', num, '/'] + den_path
        else:
            return value * num / den, ['*', num, '/'] + den_path

    # from Dimensionless to ratio of same units
    if from_unit.lower() in dictionary['Dimensionless'] and '/' in to_unit and len(to_unit.split('/')) == 2 and \
            to_unit.lower().split('/')[0].strip(' ()') == to_unit.lower().split('/')[1].strip(' ()'):
        return (lambda x: x, []) if value is None else (value, [])

    # from ratio of same units to Dimensionless
    if to_unit.lower() in dictionary['Dimensionless'] and '/' in from_unit and len(from_unit.split('/')) == 2 and \
            from_unit.lower().split('/')[0].strip(' ()') == from_unit.lower().split('/')[1].strip(' ()'):
        return (lambda x: x, []) if value is None else (value, [])

    # if inverted ratios
    if ('/' in from_unit and len(from_unit.split('/')) == 2) and ('/' in to_unit and len(to_unit.split('/')) == 2) and (
            from_unit.split('/')[0].strip() == to_unit.split('/')[1].strip()) and (
            from_unit.split('/')[1].strip() == to_unit.split('/')[0].strip()):
        if value is None:
            return lambda x: 1 / x, ['1/']
        else:
            return 1 / value, ['1/']

    # check if path is already defined in network
    conversion_path = _search_network(from_unit, to_unit)
    # return Conversion if found in network
    if conversion_path is not None:
        units_network.memory[(from_unit, to_unit)] = (_lambda_conversion(conversion_path), conversion_path)
        if value is None:
            return units_network.memory[(from_unit, to_unit)]
        else:
            return _apply_conversion(value, conversion_path), conversion_path
    else:
        return None, None
    

def generations_until_common(start, end, verbose=False, max_generations_screening=25) -> list:
    """
    Finds the number of generations requiered to get a common unit between `start` and `end`, 
    where `start` and `end` are nodes in the graph network.
    Returns a smaller number of generations to connect `start` to `end` in graph.

    Parameters
    ----------
    start: node
    end: node
    verbose: bool
        to print or not print messages.
    Returns
    -------
    shortest_path: list
    """
    from unyts.converter import _get_descendants
    max_generations_screening = unyts_parameters_.generations_limit() if max_generations_screening is None else max_generations_screening
    generations = 0
    selection = set()
    start = start.get_name() if hasattr(start, 'get_name') else start
    end = end.get_name() if hasattr(end, 'get_name') else end
    while len(selection) == 0 and generations < max_generations_screening:
        generations += 1
        start_descendants = _get_descendants(start, generations)
        end_descendants = _get_descendants(end, generations)
        selection = start_descendants.intersection(end_descendants)
    return generations if len(selection) > 0 else (np.inf if _numpy_ else 9999)


def _get_descendants(unit:str, generations=None, get_combinations=True):
    generations = unyts_parameters_.max_generations_ if generations is None else generations
    if generations == 0:
        return {unit}
    children, children_split = set(), set()
    unit_num, unit_den = None, None
    if units_network.has_node(unit):
        children = {unit}
        for g in range(generations):
            children = children.union({each for k, v in units_network.get_edges_str().items() for each in v if k in children})

    if get_combinations:
        for s in '/*':
            if s in unit and len(unit.split(s)) == 2:
                unit_num, unit_den = unit.split(s)
                break
        if unit_num is not None and unit_den is not None:
            children_num = _get_descendants(unit_num, generations)
            children_den = _get_descendants(unit_den, generations)
            children_split = {f"{str(num)}/{str(den)}"
                              for num in {unit_num}.union(children_num)
                              for den in {unit_den}.union(children_den)
                              if f"{str(num)}/{str(den)}" != unit}
        descendants = children.union(children_split)
    else:
        descendants = children
    return {unit}.union(descendants)


def _ratio_conversion_including_children(from_unit, to_unit, recursion=None, max_paths=12):
    """
    helper function of _converter function
    
    _pair_conversion_including_children will extend the search looking recursively 
    for conversion from from_unit chidren to to_unit children.
    
    the children of from_unit and to_unit are created as the ratio of combined children from 
    the numerator of the parent unit and the children of the denominator unit

    Parameters
    ----------
    from_unit : str
        unit string
    to_unit : str
        unit string
    recursion : int, optional
        limit of recursion allowed.

    Returns
    -------
    tuple
        (conversion, conversion_path)

    """
    # calculate the descendants of from_unit
    from_family = _get_descendants(from_unit)

    # calculate the descendants of to_unit
    to_family = _get_descendants(to_unit)

    # keep only intersection between families
    common = from_family.intersection(to_family)
    
    total_generations = lambda u: generations_until_common(from_unit, u) + generations_until_common(u, to_unit)
    from_to_generations = [total_generations(u) for u in common]
    min_generations = min(from_to_generations)
    common_sorted = {child: gens for gens, child in sorted(zip(from_to_generations, common))}

    # look for conversion between children  
    conversion, conversion_path = None, None
    path, shortest_path = 0, 9999
    for child in common_sorted:
        if conversion_path is not None and common_sorted[child] > min_generations:
            break
        from_child_conversion, from_child_conversion_path = _converter(None, from_unit, child, recursion=recursion)
        child_to_conversion, child_to_conversion_path = _converter(None, child, to_unit, recursion=recursion)
        if from_child_conversion is not None and child_to_conversion is not None:
            this_path_len = len([step for step in (from_child_conversion_path + child_to_conversion_path) if step != '1/'])
            if this_path_len < shortest_path:
                conversion_path = from_child_conversion_path + child_to_conversion_path
                conversion = lambda x: child_to_conversion(from_child_conversion(x))
                shortest_path = this_path_len
            if path == max_paths:
                break
        path += 1
    return conversion, conversion_path
    
    
def _converter(value, from_unit, to_unit, recursion=None):
    """
    Transform the received value (integer, float, array, series, frame, ...)
    from the units `from_unit` to the units `to_units` and report the conversion path used.

    Parameters
    ----------
    value: numeric or None
    from_unit: str
    to_unit: str

    Returns
    -------
        (conversion, conversion_path)
    """
    # avoid infinite looping, do not repeat searches
    if (from_unit, to_unit) in units_network.previous:
        return None, None
    
    # get and set recursion limit
    recursion = _get_recursion_limit(recursion)
    if unyts_parameters_.verbose_:
        logging.info(f"_converter: {recursion} remaining recursions, converting from {from_unit} to {to_unit}")
    if recursion < 0:
        return None, None

    # try to convert
    if unyts_parameters_.verbose_:
        logging.info(f"_converter: {recursion} remaining recursions, attempting direct conversion")
    conv, conv_path = _get_conversion(value, from_unit, to_unit, recursion=recursion)
    # if Conversion found
    if conv is not None:
        return conv, conv_path
    
    units_network.previous.append((from_unit, to_unit))
   
    # look for conversions of parts in ratio or product units
    if unyts_parameters_.verbose_:
        logging.info(f"_converter: {recursion} remaining recursions, attempting to convert ratio or product of units.")
    list_conversion = []
    list_conversion_path = []
    split_from = _split_unit(from_unit)
    split_to = _split_unit(to_unit)
    used_to = []
    failed = False
    for f in range(len(split_from)):
        flag = False
        if split_from[f] in '*/':
            continue
        for t in range(len(split_to)):
            if t in used_to:
                continue
            if split_to[t] in '*/':
                continue
            conv, conv_path = _get_conversion(1, split_from[f], split_to[t], recursion=recursion)
            if conv is not None:
                flag = True
                if len(list_conversion_path) > 0:
                    conv_path = [1] + conv_path
                if (f > 0 and split_from[f - 1] == '/') or (t > 0 and split_to[t - 1] == '/'):
                    conv = 1 / conv
                    conv_path = ['/'] + conv_path
                list_conversion.append(conv)
                list_conversion_path.append(conv_path)
                used_to.append(t)
                break
        if not flag:
            failed = True
            break
    if len(list_conversion) > 0 and not failed:
        conversion_factor = reduce(lambda x, y: x * y, list_conversion)
        conversion_path = [node for path in list_conversion_path for node in path]
        units_network.memory[(from_unit, to_unit)] = lambda x: x * conversion_factor, conversion_path
        if value is None:
            return units_network.memory[(from_unit, to_unit)]
        else:
            return value * conversion_factor, conversion_path

    # look for one-to-pair conversion path
    if ('/' in to_unit or '*' in to_unit) and ('/' not in from_unit and '*' not in from_unit):
        if unyts_parameters_.verbose_:
            logging.info(f"_converter: {recursion} remaining recursions, looking for one-to-pair conversion path")
        from_unit_child = _get_pair_child(from_unit)
        if from_unit_child is not None:
            base_conversion, base_conversion_path = _converter(None, from_unit, from_unit_child, recursion=recursion)
            pair_conversion, pair_conversion_path = _converter(None, from_unit_child, to_unit, recursion=recursion)
            if pair_conversion is not None and base_conversion is not None:
                conversion_path = base_conversion_path + pair_conversion_path
                conversion = lambda x: pair_conversion(base_conversion(x))
                units_network.memory[(from_unit, to_unit)] = conversion, conversion_path
                if value is None:
                    return units_network.memory[(from_unit, to_unit)]
                else:
                    return conversion(value), conversion_path

    # look for pair-to-one conversion path
    elif ('/' in from_unit or '*' in from_unit) and ('/' not in to_unit and '*' not in to_unit):
        if unyts_parameters_.verbose_:
            logging.info(f"_converter: {recursion} remaining recursions, looking for pair-to-one conversion path")
        to_unit_child = _get_pair_child(to_unit)
        if to_unit_child is not None:
            final_conversion, final_conversion_path = _converter(None, to_unit_child, to_unit, recursion=recursion)
            pair_conversion, pair_conversion_path = _converter(None, from_unit, to_unit_child, recursion=recursion)
            if pair_conversion is not None and final_conversion is not None:
                conversion_path = pair_conversion_path + final_conversion_path
                conversion = lambda x: final_conversion(pair_conversion(x))
                units_network.memory[(from_unit, to_unit)] = conversion, conversion_path
                if value is None:
                    return units_network.memory[(from_unit, to_unit)]
                else:
                    return conversion(value), conversion_path

    # look for pair to pair conversion path considering children
    if ('/' in from_unit) and ('/' in to_unit) and len(from_unit.split('/')) == 2 and len(to_unit.split('/')) == 2:
        if unyts_parameters_.verbose_:
            logging.info(f"_converter: {recursion} remaining recursions, looking for pair-to-pair conversion path considering children")
        conversion, conversion_path = _ratio_conversion_including_children(from_unit, to_unit, recursion=recursion)
        if conversion is not None:        
            units_network.memory[(from_unit, to_unit)] = conversion, conversion_path
            if value is None:
                return units_network.memory[(from_unit, to_unit)]
            else:
                return conversion(value), conversion_path

    if unyts_parameters_.verbose_:
        logging.error(f"_converter: {recursion} remaining recursions, no conversion found")
    units_network.memory[(from_unit, to_unit)] = None, None
    return None, None


def _clean_input(value: numeric, from_unit: str, to_unit: str_Empty) -> (numeric, str, str):
    """
    Helper function to preprocess the input parameters from the user
    Parameters
    ----------
    value: numeric, Unit or None
    from_unit: str, Unit or None
    to_unit: str, Unit, None or Empty

    Returns
    -------
        value: numeric
        from_unit: str
        to_unit: str
    """
    from unyts.unit_class import Unit

    if isinstance(value, Unit):
        if to_unit is Empty or to_unit == 'Empty':
            from_unit, to_unit = value.unit, from_unit
        value = value.get_value()
    if _numpy_ and hasattr(value, '__iter__') and type(value) is not np.array:
        value = np.array(value)
    elif not _numpy_ and type(value) in [list, tuple]:
        raise TypeError("Can't operate with list or tuple without NumPy. `value` can not be a list or tuple.")
    if value is not None and not isinstance(value, _numeric):
        raise ValueError("value must be numeric.")
    if isinstance(from_unit, Unit):
        from_unit = from_unit.get_unit()
    if isinstance(to_unit, Unit):
        to_unit = to_unit.get_unit()

    if isinstance(from_unit, str) and from_unit not in ('"', "'"):
        from_unit = from_unit.strip("( ')").strip('( ")').strip("'")
    elif isinstance(from_unit, str):
        from_unit = from_unit.strip("( )")
    elif from_unit is None:
        from_unit = 'None'
    elif _numpy_ and isinstance(from_unit, float) and np.isnan(from_unit):
        from_unit = 'None'
    else:
        raise TypeError(f"'from_unit' must be string, not {type(from_unit)}, like {from_unit}")

    if to_unit is Empty or to_unit == 'Empty':
        raise TypeError("convert() missing 1 required positional argument: 'to_unit'")
    if isinstance(to_unit, str) and to_unit not in ('"', "'"):
        to_unit = to_unit.strip("( ')").strip('( ")').strip("'")
    elif isinstance(to_unit, str):
        to_unit = to_unit.strip("( )")
    elif to_unit is None:
        to_unit = 'None'
    elif _numpy_ and isinstance(to_unit, float) and np.isnan(to_unit):
        to_unit = 'None'
    else:
        raise TypeError(f"'to_unit' must be string, not {type(to_unit)}, like {to_unit}")

    if unyts_parameters_.reduce_parentheses_:
        from_unit = _reduce_parentheses(from_unit)
        to_unit = _reduce_parentheses(to_unit)

    return value, from_unit, to_unit


def _clean_print_conversion_path(print_conversion_path: bool = None) -> bool:
    return unyts_parameters_.print_path_ if print_conversion_path is None else bool(print_conversion_path)


def _clean_verbose(verbose) -> bool:
    return unyts_parameters_.verbose_ if verbose is None else bool(verbose)


def _density_conversion(value: numeric, from_unit: str, to_unit: str):
    """
    Helper function to deal with conversion between weight and volume, using density.

    Parameters
    ----------
    value: numeric
    from_unit: str
    to_unit: str

    Returns
    -------
    (conv, conv_path): (numeric or lambda, list) or (None, None)
        a conversion and the conversion path, if found.
    """
    if from_unit not in uncertain_names and to_unit not in uncertain_names and \
            from_unit in dictionary['Volume'] and to_unit in dictionary['Weight']:
        density = _get_density()
        # conv1, conv_path1 = _converter(1, from_unit, 'cc')
        # conv2, conv_path2 = _converter(value, 'g', to_unit)
        this_units_density, _ = _converter(density, 'g/cm3', f"{to_unit}/{from_unit}")
        if value is None:
            # conv = lambda x: conv2(conv1(x) * density)
            conv = lambda x: x * this_units_density
        else:
            # conv = conv1 * density * conv2
            conv = value * this_units_density
        # conv_path = conv_path1 + ['*', density, '*'] + conv_path2
        conv_path = ['*', this_units_density]
    elif from_unit not in uncertain_names and to_unit not in uncertain_names and \
            to_unit in dictionary['Volume'] and from_unit in dictionary['Weight']:
        density = _get_density()
        this_units_density = _converter(density, 'g/cm3', f"{from_unit}/{to_unit}")
        # conv1, conv_path1 = _converter(value, from_unit, 'g')
        # conv2, conv_path2 = _converter(value, 'cc', to_unit)
        if value is None:
            # conv = lambda x: conv2(conv1(x) / density)
            conv = lambda x: x / this_units_density
        else:
            # conv = conv1 / density * conv2
            conv = value / this_units_density
        # conv_path = conv_path1 + ['/', density, '*'] + conv_path2
        conv_path = ['/', this_units_density]
    else:
        conv, conv_path = None, None

    return conv, conv_path


def _search_network(from_unit, to_unit, algorithm:str=None):
    """
    Searches for a path from `from_units` through the units network to `to_units`.

    Parameters
    ----------
    from_unit: str
        the starting point of the search.
    to_unit: str
        the target to be found in the network.
    algorithm: str
        the algorithm to use for searching the network, the implemented alternatives are:
        - 'BFS': Breadth-First Search algorithm, will return the shortest path searching through the entire network one level at a time.
        - 'lean_BFS': will return the shortest path searching through a slimmed network of preselected nodes.
        - 'DFS': Depth-First Search algorithm, will return the first found path searching through the network one branch at a time.
    Returns
    -------
        conversion_path: list
            the list of steps to go from `from_unit` until `to_unit`.
    """
    if algorithm is None:
        algorithm = unyts_parameters_.algorithm_
    if units_network.has_node(from_unit) and units_network.has_node(to_unit):
        if algorithm == 'BFS':
            conversion_path = BFS(units_network,
                                  units_network.get_node(from_unit),
                                  units_network.get_node(to_unit),
                                  verbose=unyts_parameters_.verbose_ and unyts_parameters_.verbose_details_ > 1)
        elif algorithm == 'lean_BFS':
            conversion_path = lean_BFS(units_network, units_network.get_node(from_unit),
                                       units_network.get_node(to_unit),
                                       verbose=unyts_parameters_.verbose_,
                                       max_generations_screening=unyts_parameters_.generations_limit())
        elif algorithm == 'DFS':
            conversion_path = DFS(units_network, units_network.get_node(from_unit),
                                  units_network.get_node(to_unit),
                                  verbose=unyts_parameters_.verbose_ and unyts_parameters_.verbose_details_ > 1,
                                  branch_depht=unyts_parameters_.generations_limit())
        else:
            raise NotImplementedError("other search algorithms different from BFS and DFS are not yet implemented.")
    else:
        conversion_path = None
    return conversion_path


def convertible(from_unit: str, to_unit: str) -> bool:
    """
    Returns True if a conversion path from `from_unit` to `to_unit` is found, otherwise returns True.

    Parameters.
    ----------
    from_unit: str, Unit or None
    to_unit: str, Unit or None

    Returns
    -------
        bool
    """
    from unyts.unit_class import Unit
    if isinstance(from_unit, Unit):
        from_unit = from_unit.get_unit()
    if isinstance(to_unit, Unit):
        to_unit = to_unit.get_unit()

    try:
        conv, conv_path = _converter(1, from_unit, to_unit)
        return False if conv is None else True
    except NoConversionFoundError:
        return False


def convert(value: numeric, from_unit: str, to_unit: str_Empty = Empty, print_conversion_path: bool = None):
    """
    Converts the received value (integer, float, array, Series, Frame, ...) from the units 'from_unit' to the units 'to_units'.

    Parameters
    ----------
    value : numeric (int, float, NumPy.array, Series, DataFrame, ...)
        the value to be converted.
    from_unit : str, Unit or None
        the units of the provided value.
    to_unit : str, Unit, None or Empty (the Empty class or the string "Empty")
        the units to convert the value.
    print_conversion_path : bool, optional
        Set to True to show the path used for Conversion. The default is False.

    Returns
    -------
    lambda_conversion : lambda
        if input value is None, or
    converted_value : int, float, array, Series, DataFrame ...
        the converted value if input value is not None
    """
    if unyts_parameters_.verbose_:
        logging.info("convert: STARTING...")
    # cleaning inputs
    value, from_unit, to_unit = _clean_input(value, from_unit, to_unit)
    print_conversion_path = _clean_print_conversion_path(print_conversion_path)
    # reset previous memory for this variable
    units_network.previous = []

    # density factor, in case of conversion from volume to mass or mass to volume
    conv, conv_path = _density_conversion(value, from_unit, to_unit)

    if conv is None:
        # regular conversion
        conv, conv_path = _converter(value, from_unit, to_unit)

    if conv is None:
        if unyts_parameters_.raise_error_:
            raise NoConversionFoundError(f"from '{from_unit}' to '{to_unit}'")
        else:
            return None

    unyts_parameters_.last_path_str = print_path(conv_path)
    if print_conversion_path:
        logging.info(f"converting from '{from_unit}' to '{to_unit}':\n {unyts_parameters_.last_path_str}")

    return conv


def convert_for_SimPandas(value: numeric, from_unit: str, to_unit: str, print_conversion_path: bool = False):
    """
    Particular implementation of `convert` function, specially set for SimDataFrames and SimSeries.
    If possible to convert the units, the returns the received value (integer, float, array, Series, Frame, ...)
    transformed from the units 'from_unit' to the units 'to_units'.
    If not possible to convert, return the original `value`.

    Parameters
    ----------
    value : Series, DataFrame
        the value to be converted.
    from_unit : str
        the units of the provided value.
    to_unit : str
        the units to convert the value.
    print_conversion_path : bool, optional
        Set to True to show the path used for Conversion. The default is False.

    Returns
    -------
    converted_value : Series, DataFrame
        the converted value if input value is not None
    """
    conv = None
    print_conversion_path = bool(print_conversion_path)
    if convertible(from_unit, to_unit):
        conv, conv_path = _converter(value, from_unit, to_unit)
    if print_conversion_path and conv is not None:
        logging.info("converting from '{from_unit}' to '{to_unit}':\n {print_path(conv_path)}")
    elif print_conversion_path and conv is not None:
        logging.warning("conversion not found, returning original values.")
    return value if conv is None else conv



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 11 10:38:47 2024

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.3'
__release__ = 20240811
__all__ = ['start_gui']

import logging
import tkinter as tk
from tkinter import filedialog
from tkinter import simpledialog
from time import process_time
from tkinter import ttk
from stringthings import get_number, is_numeric
from unyts import __version__ as unyts_version
from .converter import convert
from .dictionaries import _all_units
from .errors import NoConversionFoundError, NoFVFError
from .database import save_memory, load_memory, clean_memory, delete_cache, units_network  # set_fvf
from .parameters import unyts_parameters_  # set_density
import pathlib, os
import webbrowser


class UnytsApp(tk.Frame):
    _all_units_str = _all_units()

    def __init__(self, master=None):
        super().__init__(master)
        self.pack()

        ttk.Label(text="Unyts converter", font=(None, 16, 'bold'), padding=5).pack()

        frame = ttk.Frame(padding=5)
        self.frame = frame
        self.frame.pack()

        ttk.Label(frame, text="unit").grid(column=1, row=0, pady=3)
        ttk.Label(frame, text="value").grid(column=2, row=0, pady=3)

        ttk.Label(frame, text="from").grid(column=0, row=1)
        self.from_unit_val = tk.StringVar()
        self.from_unit = ttk.Entry(frame, textvariable=self.from_unit_val)
        self.from_unit.grid(column=1, row=1, pady=3, padx=1)
        self.from_unit.bind('<Key-Return>', self._calculate)
        self.from_value_val = tk.StringVar()
        self.from_value = ttk.Entry(frame, textvariable=self.from_value_val)
        self.from_value.grid(column=2, row=1, pady=3, padx=1)
        self.from_value.bind('<Key-Return>', self._calculate)

        ttk.Label(frame, text="to").grid(column=0, row=2)
        self.to_unit_val = tk.StringVar()
        self.to_unit = ttk.Entry(frame, textvariable=self.to_unit_val)
        self.to_unit.grid(column=1, row=2, pady=3, padx=1)
        self.to_unit.bind('<Key-Return>', self._calculate)
        self.to_value_val = tk.StringVar()
        self.to_value = ttk.Entry(frame, textvariable=self.to_value_val)
        self.to_value.grid(column=2, row=2, pady=3, padx=1)
        self.to_value.bind('<Key-Return>', self._rcalculate)

        self.button_text = tk.StringVar()
        self.button_text.set("convert")
        self.convert_button = ttk.Button(frame, textvariable=self.button_text)
        self.convert_button.grid(columnspan=3, row=3, pady=7)
        self.convert_button.bind('<ButtonRelease>', self._calculate)
        self.frame.bind('<Motion>', lambda x: self.button_text.set("convert"))

        self.result_str = tk.StringVar()
        self.result_str.set("")
        self.result_msg = ttk.Label(frame, textvariable=self.result_str)
        self.result_msg.config(foreground='#5f5f5f')
        self.result_msg.grid(column=0, columnspan=3, row=4)

        # conversion path marquee
        self._first_char = tk.IntVar()
        self._first_char.set(0)
        self._msg_len = tk.IntVar()
        self._msg_len.set(len(unyts_parameters_.last_path_str))
        self._msg_dir = tk.BooleanVar()
        self._msg_dir.set(True)
        self._max_len = 53
        self._path_marquee = None

        path_frame = ttk.Frame(padding=0)
        self.path = path_frame
        if unyts_parameters_.print_path_:
            self.path.pack()
        self.path_str = tk.StringVar()
        self.path_str.set("")
        self.path_msg = ttk.Label(path_frame, textvariable=self.path_str)
        self.path_msg.grid(column=0, row=0)

        # # user input row
        # user_input = ttk.Frame(padding=0)
        # self.input = user_input
        # self.input_label_str = tk.StringVar()
        # self.input_label = ttk.Label(textvariable=self.input_label_str)
        # # self.input_label.grid(column=0, row=0)
        # self.input_value_str = tk.StringVar()
        # self.input_value = ttk.Entry(user_input, textvariable=self.input_value_str)
        # # self.input_value.grid(column=1, row=0)
        # self.input_button = ttk.Button(user_input, text='set')
        # # self.input_button.grid(columnspan=3, row=0)
        # self.input_button.bind('<ButtonRelease>', self._get_input)

    def _validate_from_units(self, *args):
        if self.from_unit_val.get() not in UnytsApp._all_units_str and not is_numeric(self.from_unit_val.get()):
            return False
        else:
            return True

    def _validate_to_units(self, *args):
        if self.to_unit_val.get() not in UnytsApp._all_units_str and not is_numeric(self.to_unit_val.get()):
            return False
        else:
            return True

    def _get_from(self):
        from_value = self.from_value_val.get().strip().replace("no conversion found!", "").replace("wrong number format", "")
        from_unit = self.from_unit_val.get().strip()
        if len(from_value) == 0 and len(from_unit) > 0 and not is_numeric(from_unit):
            pass  # is OK
        elif len(from_unit) == 0 and len(from_value) > 0 and not is_numeric(from_value):
            from_value, from_unit = from_unit, from_value
        elif is_numeric(from_value) and not is_numeric(from_unit):
            from_value = get_number(from_value)
        elif not is_numeric(from_value) and is_numeric(from_unit):
            from_value, from_unit = get_number(from_unit), from_value
        elif not is_numeric(from_value) and not is_numeric(from_unit):
            self.from_value_val.set("wrong number format")
            from_value, from_unit = None, None
        self.from_unit_val.set(from_unit)
        self.from_value_val.set(from_value)
        return from_unit, from_value

    def _get_to(self):
        to_value = self.to_value_val.get().strip().replace("no conversion found!", "").replace("wrong number format", "")
        to_unit = self.to_unit_val.get().strip()
        if len(to_value) == 0 and len(to_unit) > 0 and not is_numeric(to_unit):
            pass  # is OK
        elif len(to_unit) == 0 and len(to_value) > 0 and not is_numeric(to_value):
            to_value, to_unit = to_unit, to_value
        elif is_numeric(to_value) and not is_numeric(to_unit):
            to_value = get_number(to_value)
        elif not is_numeric(to_value) and is_numeric(to_unit):
            to_value, to_unit = get_number(to_unit), to_value
        elif not is_numeric(to_value) and not is_numeric(to_unit):
            self.to_value_val.set("wrong number format")
            to_value, to_unit = None, None
        self.to_unit_val.set(to_unit)
        self.to_value_val.set(to_value)
        return to_unit, to_value

    def _calculate(self, *args):
        self._display_path(" ")
        start_time = process_time()
        from_unit, from_value = self._get_from()
        to_unit, to_value = self._get_to()
        if from_unit is None and from_value is None and to_unit is None and to_value is None:
            self.from_value_val.set("")
            self.from_unit_val.set("")
            self.to_value_val.set("")
            self.to_unit_val.set("")
            self.button_text.set("convert")
            return
        if from_value == "" and is_numeric(to_value):
            return self._rcalculate()
        try:
            to_value = convert(from_value, from_unit, to_unit,
                               print_conversion_path=unyts_parameters_.print_path_)
            self.to_value_val.set(str(to_value))
            end_time = process_time()
            self._calculation_time(start_time, end_time)
            self._display_path()
        except NoConversionFoundError:
            self.to_value_val.set("")
            end_time = process_time()
            self._calculation_time(start_time, end_time)
            self._display_path("no conversion found!")
            return
        except NoFVFError:
            self._display_path(error="Please set FVF constant from Options menu.")
            return

    def _rcalculate(self, *args):
        self._display_path(" ")
        start_time = process_time()
        from_unit, from_value = self._get_to()
        to_unit, to_value = self._get_from()
        if from_unit is None and from_value is None and to_unit is None and to_value is None:
            self.from_value_val.set("")
            self.from_unit_val.set("")
            self.to_value_val.set("")
            self.to_unit_val.set("")
            self.button_text.set("convert")
            return
        if from_value == "" and is_numeric(to_value):
            return self._calculate()
        try:
            to_value = convert(from_value, from_unit, to_unit,
                               print_conversion_path=unyts_parameters_.print_path_)
            self.from_value_val.set(str(to_value))
            end_time = process_time()
            self._calculation_time(start_time, end_time)
            self._display_path()
        except NoConversionFoundError:
            self.from_value_val.set("")
            end_time = process_time()
            self._calculation_time(start_time, end_time)
            self._display_path("no conversion found!")
            return
        except NoFVFError:
            self._display_path(error="Please set FVF constant from Options menu.")
            return

    def _calculation_time(self, start_time, end_time):
        total_time = end_time - start_time
        if total_time < 1:
            total_time = round(total_time * 1000, 0)
            time_units = 'milliseconds'
        elif total_time > 60:  # 1 minute
            minutes = total_time // 60
            seconds = round(total_time - minutes, 0)
            total_time = f"{minutes}:{seconds}"
            time_units = 'minutes:seconds'
        elif total_time > 3600:  # 1 hour
            hours = total_time // 3600
            minutes = round((total_time - hours) / 60, 0)
            total_time = f"{hours}:{minutes}"
            time_units = 'hours:minutes'
        else:
            total_time = round(total_time, 2)
            time_units = 'seconds'
        total_time = f"{total_time} {time_units}"
        self.result_str.set(total_time)

    def _partial_msg(self):
        if len(unyts_parameters_.last_path_str) < self._max_len:
            self.path.after_cancel(self._path_marquee)
            self.path_str.set(unyts_parameters_.last_path_str)
            return
        _last_char = self._first_char.get() + self._max_len
        speed = 100
        msg = unyts_parameters_.last_path_str[self._first_char.get(): _last_char]
        self.path_str.set(msg)
        if self._msg_dir.get():
            if _last_char < self._msg_len.get():
                self._first_char.set(self._first_char.get() + 1)
            else:
                self._msg_dir.set(False)
                speed = 2000
        else:
            if self._first_char.get() > 0:
                self._first_char.set(self._first_char.get() - 1)
            else:
                self._msg_dir.set(True)
                speed = 2000
        self._path_marquee = self.path.after(speed, self._partial_msg)

    def _display_path(self, error:str=""):
        if bool(error):
            if self._path_marquee is not None:
                self.path.after_cancel(self._path_marquee)
            self.path_str.set(error)
            self.path_msg.config(foreground='red')
            return
        self.path_msg.config(foreground='#3d3d3d')
        if unyts_parameters_.last_path_str != "" and unyts_parameters_.print_path_:
            if len(unyts_parameters_.last_path_str) < self._max_len:
                self.path_str.set(unyts_parameters_.last_path_str)
            else:
                self._msg_len.set(len(unyts_parameters_.last_path_str))
                self.path_str.set(unyts_parameters_.last_path_str[:self._max_len])
                self._path_marquee = self.path.after(2000, self._partial_msg)
        else:
            self.path_str.set("")

    def _get_input(self):
        self.input_value_str = self.input_value.get().strip()


def start_gui():
    unyts_parameters_.gui = True
    def close_gui():
        logging.info("INFO:shutting down Unyts.")
        if unyts_parameters_.cache_:
            logging.info("saving memory...")
            save_memory()
        root.destroy()
    if unyts_parameters_.memory_:
        load_memory()
    else:
        delete_cache()
    def open_help():
        webbrowser.open('https://github.com/ayaranitram/unyts/blob/master/unyts_demo.ipynb')
    def open_git():
        webbrowser.open('https://github.com/ayaranitram/unyts')
    def print_path():
        unyts_parameters_.print_path()
        if unyts_parameters_.print_path_:
            unyts_gui.path.pack()
            unyts_gui._display_path()
        else:
            unyts_gui._display_path(" ")
            unyts_gui.path.pack_forget()
    def _set_bfs():
        unyts_parameters_.set_algorithm('BFS')
        _bfs_.set(True)
        _lean_bfs_.set(False)
        _dfs_.set(False)
    def _set_lean_bfs():
        unyts_parameters_.set_algorithm('lean_BFS')
        _bfs_.set(False)
        _lean_bfs_.set(True)
        _dfs_.set(False)
    def _set_dfs():
        unyts_parameters_.set_algorithm('DFS')
        _bfs_.set(False)
        _lean_bfs_.set(False)
        _dfs_.set(True)
    def _set_fvf():
        if unyts_parameters_.fvf_ is not None:
            current_value = f"{unyts_parameters_.fvf_} vol/vol"
            prev_fvf_ = unyts_parameters_.fvf_
        else:
            current_value = None
            prev_fvf_ = None
        fvf = None
        while fvf is None:
            user_input = simpledialog.askstring("formation volume factor",
                                     "Enter FVF in vol/vol or type the value followed by the units, like 1.5 rb/MMscf",
                                                initialvalue=current_value)
            if user_input is None or len(user_input) == 0:
                fvf = False
            elif len(user_input.split()) == 1:
                try:
                    user_input = get_number(user_input)
                    if user_input > 0:
                        fvf = user_input
                except:
                    pass
            elif len(user_input.split()) == 2:
                user_input, unit = user_input.split()
                if unit.lower() == 'vol/vol':
                    unit = 'sm3/sm3'
                try:
                    user_input = get_number(user_input)
                    if user_input > 0:
                        try:
                            units_network.set_fvf(1)
                            unyts_parameters_.fvf_ = 1
                            fvf = convert(user_input, unit, 'sm3/sm3')
                        except NoConversionFoundError:
                            unyts_gui._display_path(error="FVF units {unit} are not valid.")
                            units_network.set_fvf(prev_fvf_)
                            unyts_parameters_.fvf_ = prev_fvf_
                except:
                    pass
        units_network.set_fvf(fvf)
        unyts_parameters_.fvf_ = fvf
        unyts_parameters_.save_params()
        if user_input == fvf:
            unyts_parameters_.last_path_str = f"FVF set as {fvf} vol/vol."
        else:
            unyts_parameters_.last_path_str = f"FVF {user_input} {unit} converted to {fvf} vol/vol"
        unyts_gui._display_path()
    def _set_density():
        if unyts_parameters_.density_ is not None:
            current_value = f"{unyts_parameters_.density_} g/cm³"
        else:
            current_value = None
        density = None
        while density is None:
            user_input = simpledialog.askstring("constant density to convert between volume and mass",
                                     "Enter density in g/cm³ or type the value followed by the units, like: 997 kg/m³",
                                                initialvalue=current_value)
            print(user_input)
            if user_input is None or len(user_input) == 0:
                density = False
            elif len(user_input.split()) == 1:
                try:
                    user_input = get_number(user_input)
                    if user_input > 0:
                        density = user_input
                except:
                    pass
            elif len(user_input.split()) == 2:
                user_input, unit = user_input.split()
                try:
                    user_input = get_number(user_input)
                    if user_input > 0:
                        try:
                            density = convert(user_input, unit, 'g/cm3')
                        except NoConversionFoundError:
                            unyts_gui._display_path(error=f"density units {unit} are not valid.")
                except:
                    pass
        unyts_parameters_.density_ = density
        unyts_parameters_.save_params()
        if user_input == density:
            unyts_parameters_.last_path_str = f"density set as {density} g/cm³."
        else:
            unyts_parameters_.last_path_str = f"density {user_input} {unit} converted to {density} g/cm³"
        unyts_gui._display_path()

    def export_memory():
        cache_path = filedialog.asksaveasfilename(
            title='Save',
            filetypes=(('Cache files', '*.cache'),('All Files', '*.*')),
            initialfile='unyts_memory.cache'
        )
        if cache_path:
            try:
                save_memory(cache_path)
            except:
                logging.error(f"Failed to save memory to file {cache_path}")
        else:
            pass
    def load_memory_file():
        cache_path = filedialog.askopenfilename(
            title='Load',
            filetypes=(('Cache files', '*.cache'),('All Files', '*.*')))
        if cache_path:
            if os.path.isfile(cache_path):
                try:
                    load_memory(cache_path)
                    unyts_gui._display_path()
                except:
                    msg = f"Not able to load memory from file {cache_path}"
                    logging.error(msg)
                    unyts_gui._display_path(error=msg)
            else:
                msg = f"The file {cache_path} doesn't exists."
                logging.error(msg)
                unyts_gui._display_path(error=msg)
        else:
            pass

    def set_cache_folder():
        cache_path = unyts_parameters_.get_user_folder()
        cache_path = filedialog.askdirectory(initialdir=cache_path, title='Cache Folder')
        if cache_path:
            if os.path.isdir(cache_path):
                unyts_parameters_.set_user_folder(cache_path)
            else:
                msg = f"The file {cache_path} doesn't exists."
                logging.error(msg)
                unyts_gui._display_path(error=msg)
        else:
            pass

    logging.info("starting Unyts GUI...")
    w, h = 325, 215
    root = tk.Tk(screenName='Unyts')
    root.geometry(f"{w}x{h}")
    root.maxsize(w, h)
    root.minsize(w, h)
    root.resizable(False, False)
    icon_file = 'unyts_icon.ico'
    current_dir = pathlib.Path(__file__).parent.resolve()
    icon_path = os.path.join(current_dir, icon_file)
    root.iconbitmap(default=icon_path)

    # variables
    _cache_ = tk.BooleanVar()
    _cache_.set(unyts_parameters_.cache_)
    _verbosity_ = tk.BooleanVar()
    _verbosity_.set(unyts_parameters_.verbose_)
    _print_path_ = tk.BooleanVar()
    _print_path_.set(unyts_parameters_.print_path_)
    _bfs_, _lean_bfs_, _dfs_ = tk.BooleanVar(), tk.BooleanVar(), tk.BooleanVar()
    _bfs_.set(unyts_parameters_.algorithm_ == 'BFS')
    _lean_bfs_.set(unyts_parameters_.algorithm_ == 'lean_BFS')
    _dfs_.set(unyts_parameters_.algorithm_ == 'DFS')

    # setting menu
    root.option_add('*tearOff', False)
    unyts_menu = tk.Menu(root)
    root.config(menu=unyts_menu)
    # File menu
    file_menu = tk.Menu(unyts_menu)
    unyts_menu.add_cascade(label='File', menu=file_menu)
    # File menu > Memory menu
    memory_menu = tk.Menu(unyts_menu)
    memory_menu.add_command(label='Save memory', command=save_memory)
    memory_menu.add_command(label='Clean memory', command=clean_memory)
    memory_menu.add_command(label='Reload memory', command=load_memory)
    memory_menu.add_command(label='Export memory', command=export_memory)
    memory_menu.add_command(label='Import memory', command=load_memory_file)
    file_menu.add_cascade(label='Conversions memory', menu=memory_menu)
    file_menu.add_command(label='Delete cache files', command=delete_cache)
    file_menu.add_separator()
    file_menu.add_command(label='Exit', command=close_gui)
    # Options menu
    options_menu = tk.Menu(unyts_menu)
    unyts_menu.add_cascade(label='Options', menu=options_menu)
    # options_menu.add_command(label="Set FVF (in CMD)", command=set_fvf)
    options_menu.add_command(label="Set FVF", command=_set_fvf)
    options_menu.add_command(label="Set density", command=_set_density)
    #options_menu.add_command(label="Set density (in CMD)", command=set_density)
    options_menu.add_separator()
    options_menu.add_checkbutton(label='Rebuild on next start', variable=unyts_parameters_.reload_,
                                 command=unyts_parameters_.reload_next_time)
    options_menu.add_checkbutton(label='Show conversion path', variable=_print_path_, command=print_path)
    options_menu.add_checkbutton(label='Verbose', variable=_verbosity_, command=unyts_parameters_.verbose)
    options_menu.add_checkbutton(label='Use cache', variable=_cache_, command=unyts_parameters_.cache)
    options_menu.add_separator()
    options_menu.add_command(label='Set cache folder', command=set_cache_folder)
    # Search algorith menu
    search_menu = tk.Menu(unyts_menu)
    search_menu.add_checkbutton(label='BFS', variable=_bfs_, command=_set_bfs)
    search_menu.add_checkbutton(label='lean BFS', variable=_lean_bfs_, command=_set_lean_bfs)
    search_menu.add_checkbutton(label='DFS', variable=_dfs_, command=_set_dfs)
    options_menu.add_cascade(label='Search algorithm', menu=search_menu)
    # help menu
    help_menu = tk.Menu(unyts_menu)
    unyts_menu.add_cascade(label='Help', menu=help_menu)
    help_menu.add_command(label='Help', command=open_help)
    help_menu.add_command(label='GitHub', command=open_git)
    help_menu.add_command(label=f"Version {unyts_version}")

    unyts_gui = UnytsApp()
    unyts_gui.master.title("Unyts converter")
    root.protocol("WM_DELETE_WINDOW", close_gui)
    unyts_gui.mainloop()



"""
Created on Sat Dec 30 22:27:33 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.0.0'
__release__ = 20231230
__all__ = ['Empty', 'str_Empty', 'str_Empty_']

from typing import Union


class EmptyType(type):
    def __repr__(self):
        return "Empty"


class Empty(object, metaclass=EmptyType):
    """
    A class to specify an "Empty" (not Null nor None) value.
    """


str_Empty = Union[str, Empty]
str_Empty_ = (str, Empty)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 18:24:20 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.9.8'
__release__ = 20240823
__all__ = ['units', 'convert', 'convertible', 'Unit', 'is_Unit', 'valid_unit',
           'set_unit', 'set_conversion', 'set_density', 'get_density'
           'save', 'start_gui', 'set_fvf', 'set_algorithm']

from .parameters import unyts_parameters_, print_path, reload, raise_error, cache, set_density, get_density, recursion_limit, verbose, set_algorithm, get_algorithm
from .database import network_to_frame, save_memory, load_memory, clean_memory, set_fvf, get_fvf
from .units.define import units
from .converter import convert, convertible
from .Empty import Empty
from .unit_class import Unit, is_Unit, valid_unit
from .units.custom import set_unit, set_conversion
from .gui import start_gui


if unyts_parameters_.show_version_:
    print(f"loaded unyts version {str(__version__)}")
    unyts_parameters_.show_version_ = False
    unyts_parameters_.save_params()

from .unitary import *


def save(path=None) -> None:
    """
    Will save the current state of searches memory to a cache file that will be loaded next time unyts is started.

    Returns
    -------
    None
    """
    save_memory(path)


def load(path=None) -> None:
    """
    Will load the saved cache searches into current memory of the converter.

    Returns
    -------
    None
    """
    load_memory(path)



"""
Created on Mon Jan  30 08:14:12 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

import unyts
from unyts import Unit
from unyts.units.geometry import Volume
from unyts.units.force import Weight
from unyts.errors import WrongUnitsError

oz = unyts.units(1, 'oz')
assert type(oz) is Unit
oz_vol = oz.to('ml')
assert type(oz_vol) is Volume
oz_wei = oz.to('g')
assert type(oz_wei) is Weight

assert type(oz_vol.to('ml').to('oz')) is Volume
assert type(oz_wei.to('g').to('oz')) is Weight

try:
    oz_vol.to('ml').to('oz').to('g')
    raise WrongUnitsError('oz volume converted to gram')
except WrongUnitsError:
    pass

try:
    oz_wei.to('g').to('oz').to('ml')
    raise WrongUnitsError('oz volume converted to millilitre')
except WrongUnitsError:
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 26 21:45:34 2023

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['Date']

from ..dictionaries import dictionary as _dictionary
from ..unit_class import Unit
from .time import Time
from ..helpers.common_classes import unit_or_str, numeric
from numpy import datetime64, timedelta64


class Date(Unit):
    class_units = _dictionary['Date']
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value, units='date', name=None):
        self.value = datetime64(value)
        self.name = 'date' if name is None else name
        self.kind = Date

    def __add__(self, other):
        if isinstance(other, Time):
            return Date(self.value + timedelta64(other.value, other.unit))

    def __sub__(self, other):
        if isinstance(other, Time):
            return Date(self.value - timedelta64(other.value, other.unit))

    def year(self):
        return self.value.astype(object).year

    def month(self):
        return self.value.astype(object).month

    def day(self):
        return self.value.astype(object).day



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.6.4'
__release__ = 20240807
__all__ = ['Unit', 'is_Unit']

import logging
from .errors import WrongUnitsError, WrongValueError, NoConversionFoundError
from .operations import unit_product as _unit_product, unit_division as _unit_division, \
    unit_base_power as _unit_base_power
from .helpers.unit_string_tools import reduce_units as _reduce_units
from .converter import convert as _convert, convertible as _convertible
from .dictionaries import _all_units
from .parameters import unyts_parameters_
from numbers import Number
from typing import Union

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

try:
    import numpy as np
    from numpy import ndarray, int64, float64, int32, float32
    _numpy_ = True
except ModuleNotFoundError:
    _numpy_ = False
try:
    from pandas import Series, DataFrame
    _pandas_ = True
except ModuleNotFoundError:
    _pandas_ = False

if _numpy_ and _pandas_:
    _number = (int, float, complex, int32, int64, float32, float64)
    number = Union[int, float, complex, int32, int64, float32, float64]
    _numeric = (int, float, complex, int32, int64, float32, float64, ndarray, Series, DataFrame)
    numeric = Union[int, float, complex, int32, int64, float32, float64, ndarray, Series, DataFrame]
    _array_like = (ndarray, Series, DataFrame)
    array_like = Union[ndarray, Series, DataFrame]
elif _numpy_:
    _number = (int, float, complex, int32, int64, float32, float64)
    number = Union[int, float, complex, int32, int64, float32, float64]
    _numeric = (int, float, complex, ndarray, int32, int64, float32, float64)
    numeric = Union[int, float, complex, ndarray, int32, int64, float32, float64]
    _array_like = (ndarray, )
    array_like = Union[ndarray]
elif _pandas_:
    _number = (int, float, complex)
    number = Union[int, float, complex]
    _numeric = (int, float, complex, Series, DataFrame)
    numeric = Union[int, float, complex, Series, DataFrame]
    _array_like = (Series, DataFrame)
    array_like = Union[Series, DataFrame]
else:
    _number = (int, float, complex)
    number = Union[int, float, complex]
    _numeric = (int, float, complex)
    numeric = Union[int, float, complex]
    _array_like = (None, )
    array_like = None


class UnytType(type):
    def __repr__(self):
        return self.__name__


class Unit(object, metaclass=UnytType):
    """
    A class to cope with values associated to units, its arithmetic and logic
    operations and conversions.
    """
    class_units = []
    _all_units_str = _all_units()
    __slots__ = ('unit', 'value', 'name', 'kind')

    def __init__(self, value: numeric, unit=None, name=None):
        if isinstance(value, Unit):
            if unit is None:
                value, unit = value.value, value.unit
            else:
                value = value.value
        if unit is None:
            unit = 'dimensionless'
        elif type(unit) is str and unit.strip() in Unit._all_units_str:
            self.unit = unit.strip()
        elif unyts_parameters_.raise_error:
            raise WrongUnitsError("'" + str(unit) + "' is not a valid units name.")
        elif not unyts_parameters_.raise_error:
            logging.error("'" + str(unit) + "' is not a valid units name.")
        self.value = self.check_value(value)
        self.unit = unit
        self.name = 'unit' if name is None else name
        self.kind = Unit

    def __call__(self) -> numeric:
        return self.value

    def __repr__(self) -> str:
        return str(self.value) + '_' + str(self.unit)

    def __str__(self) -> str:
        if self.unit is None or len(str(self.unit).strip()) == 0:
            return str(self.value)
        else:
            return str(self.value) + '_' + str(self.unit)

    @property
    def dtype(self):
        if hasattr(self.values, 'dtype'):
            return self.values.dtype
        elif _numpy_:
            if type(self.value) is int:
                return np.dtype(int)
            elif type(self.value) is float:
                return np.dtype(float)
            elif type(self.value) is complex:
                return np.dtype(complex)
        else:
            raise NotImplementedError("dtype not implemented without NumPy.")

    def convert(self, new_unit: str):
        if type(new_unit) is not str and hasattr(new_unit, 'units') and type(new_unit.units) is str:
            new_unit = new_unit.units
        elif type(new_unit) is not str:
            raise TypeError("`new_units` should be string.")
        if not _convertible(self.unit, new_unit):
            from .units.custom import UserUnits
            if self.kind is UserUnits:
                logging.warning("'" + str(new_unit) + "' is not valid units for " + str(type(self)) + ".")
                return self
            else:
                raise WrongUnitsError("'" + str(new_unit) + "' is not valid units for " + str(type(self)) + ".")
        if type(self) is Unit:
            from .units.define import units
            return units(_convert(self.value, self.unit, new_unit), new_unit)
        else:
            return self.kind(_convert(self.value, self.unit, new_unit), new_unit)

    def to(self, new_unit):
        return self.convert(new_unit)

    def __neg__(self):
        return self.kind(self.value.__neg__(), self.unit)

    def __bool__(self):
        from .units.custom import UserUnits
        from .units.unitless import Dimensionless, Percentage
        if self.kind in (Dimensionless, Percentage, UserUnits):
            return False
        return True

    def __abs__(self):
        return self.kind(abs(self.value), self.unit)

    def __add__(self, other):
        from .units.unitless import Dimensionless, Percentage
        if isinstance(other, Unit):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value + other.value, self.unit)
                elif _convertible(other.unit, self.unit):
                    return self.kind(self.value + _convert(other.value, other.unit, self.unit), self.unit)
                elif _convertible(self.unit, other.unit):
                    return self.kind(other.value + _convert(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return self.kind(self.value + other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value + other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value + other.value, self.unit)
            else:  # add different units
                return self, other
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self + each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is Percentage:
            return self.kind((self.value + other) * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value + other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__radd__(self)
        else:
            raise NotImplementedError(f"Addition of {type(self)} and {type(other)} not implemented.")

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value * other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value * other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value * other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value * other.value, _unit_product(self.unit, other.unit))
            elif _convertible(other.unit, self.unit):
                return units(self.value * _convert(other.value, other.unit, self.unit),
                             _unit_product(self.unit, other.unit))
            elif _convertible(self.unit, other.unit):
                return units(other.value * _convert(self.value, self.unit, other.unit),
                             _unit_product(other.unit, self.unit))
            elif _convertible(_unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]):
                factor = _convert(1, _unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]) ** _unit_base_power(other.unit)[1]
                return units(self.value * other.value * factor, _unit_product(self.unit, other.unit))
            else:
                return units(self.value * other.value, _unit_product(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self * each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value * other * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value * other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rmul__(self)
        else:
            raise NotImplementedError(f"Product of {type(self)} and {type(other)} not implemented.")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value ** other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value ** other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value ** other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
            elif _convertible(other.unit, self.unit):
                return units(self.value ** _convert(other.value, other.unit, self.unit), self.unit + '^' + self.unit)
            elif _convertible(_unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]):
                factor = _convert(1, _unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]) ** _unit_base_power(other.unit)[1]
                return units(self.value ** (other.value * factor), self.unit + '^' + other.unit)
            else:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self ** each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind((self.value ** other) * 100, self.unit)
        elif isinstance(other, _numeric):
            pow_units = _unit_base_power(self.unit)[1] * other
            if pow_units == 0:
                pow_units = 'Dimensionless'
            elif pow_units == 1:
                pow_units = _unit_base_power(self.unit)[0]
            else:
                pow_units = _unit_base_power(self.unit)[0] + str(pow_units)
            return units(self.value ** other, pow_units)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rpow__(self)
        else:
            raise NotImplementedError(f"Power of {type(self)} to {type(other)} not implemented.")

    def __rpow__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if self.kind is Percentage:
            return units((other ** self.value), 'Dimensionless')
        elif self.kind is Dimensionless:
            return units(other ** self.value, self.unit)
        else:
            raise TypeError(f"unsupported operand type(s) for ** or pow(): '{type(other)}' and '{type(self)}.")

    def __sub__(self, other):
        # return self.__add__(other * -1)
        from .units.unitless import Dimensionless, Percentage
        if isinstance(other, Unit):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value - other.value, self.unit)
                elif _convertible(other.unit, self.unit):
                    return self.kind(self.value - _convert(other.value, other.unit, self.unit), self.unit)
                elif _convertible(self.unit, other.unit):
                    return self.kind(other.value - _convert(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return self.kind(self.value - other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value - other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value - other.value, self.unit)
            else:  # add different units
                return self, other
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self - each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is Percentage:
            return self.kind((self.value - other) * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value - other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rsub__(self)
        else:
            raise NotImplementedError(f"Subtraction of {type(self)} and {type(other)} not implemented.")

    def __rsub__(self, other):
        return self.__neg__() + other

    def __truediv__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value / other.value, '1/' + other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value / other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value / other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value / other.value, _unit_division(self.unit, other.unit))
            elif _convertible(other.unit, self.unit):
                return units(self.value / _convert(other.value, other.unit, self.unit),
                             _unit_division(self.unit, other.unit))
            elif _convertible(_unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]):
                factor = _convert(1, _unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]) ** _unit_base_power(other.unit)[1]
                return units(self.value / (other.value * factor), _unit_division(self.unit, other.unit))
            else:
                return units(self.value / other.value, _unit_division(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self / each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value / other * 100, self.unit)
        elif isinstance(other, _numeric):
            return units(self.value / other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rtruediv__(self)
        else:
            raise NotImplementedError(f"Division of {type(self)} by {type(other)} not implemented.")

    def __rtruediv__(self, other):
        from .units.define import units
        from .units.unitless import Dimensionless, Percentage
        if self.kind is Percentage:
            return units(other / self.value * 100, self.unit)
        elif self.kind is Dimensionless:
            return units(other / self.value, self.unit)
        elif isinstance(other, _numeric):
            return units(other / self.value, self.unit + '-1')
        else:
            raise NotImplementedError(f"Division of {type(self)} by {type(other)} not implemented.")

    def __floordiv__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value // other.value, '1/' + other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value // other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value // other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value // other.value, _unit_division(self.unit, other.unit))
            elif _convertible(other.unit, self.unit):
                return units(self.value // _convert(other.value, other.unit, self.unit),
                             _unit_division(self.unit, other.unit))
            elif _convertible(_unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]):
                factor = _convert(1, _unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]) ** _unit_base_power(other.unit)[1]
                return units(self.value // (other.value * factor), _unit_division(self.unit, other.unit))
            else:
                return units(self.value // other.value, _unit_division(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self // each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value // other * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value // other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rfloordiv__(self)
        else:
            raise NotImplementedError(f"Division of {type(self)} by {type(other)} not implemented.")

    def __rfloordiv__(self, other):
        from .units.define import units
        from .units.unitless import Dimensionless, Percentage
        if self.kind is Percentage:
            return units(other // self.value * 100, self.unit)
        elif self.kind is Dimensionless:
            return units(other // self.value, self.unit)
        elif isinstance(other, _numeric):
            return units(other // self.value, self.unit + '-1')
        else:
            raise NotImplementedError(f"Division of {type(self)} by {type(other)} not implemented.")

    def __matmul__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value / other.value, '1/' + other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value / other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value / other.value, self.unit)
            else:
                return units(self.value / other.value, _reduce_units(self.unit + '/' + other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self / each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value / other * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value / other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rmatmul__(self)
        else:
            raise NotImplementedError(f"Division of {type(self)} by {type(other)} not implemented.")

    def __mod__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if isinstance(other, Unit):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value % other.value, self.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value % other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value % other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value % other.value, self.unit)
            elif _convertible(other.unit, self.unit):
                return units(self.value % _convert(other.value, other.unit, self.unit), self.unit)
            elif _convertible(_unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]):
                factor = _convert(1, _unit_base_power(other.unit)[0], _unit_base_power(self.unit)[0]) ** _unit_base_power(other.unit)[1]
                return units(self.value % (other.value * factor), self.unit)
            else:
                return units(self.value % other.value, self.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self % each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value % other * 100, self.unit)
        elif isinstance(other, _numeric):
            return self.kind(self.value % other, self.unit)
        elif hasattr(other, 'type') and other.type in ('SimSeries', 'SimDataFrame'):
            return other.__rmod__(self)
        else:
            raise NotImplementedError(f"Module of {type(self)} when divided by {type(other)} not implemented.")

    def __lt__(self, other) -> bool:
        if not isinstance(other, Unit):
            return self.value < other
        elif type(self) == type(other):
            return self.value < other.convert(self.unit).value
        else:
            msg = f"'<' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __le__(self, other) -> bool:
        if not isinstance(other, Unit):
            return self.value <= other
        elif type(self) == type(other):
            return self.value <= other.convert(self.unit).value
        else:
            msg = f"'<=' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Unit):
            return self.value == other
        elif type(self) is type(other):
            return self.value == other.convert(self.unit).value
        else:
            msg = f"'==' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __ne__(self, other):
        if not isinstance(other, Unit):
            return self.value != other
        elif type(self) is type(other):
            return self.value != other.convert(self.unit).value
        else:
            msg = f"'!=' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __ge__(self, other):
        if not isinstance(other, Unit):
            return self.value >= other
        elif type(self) is type(other):
            return self.value >= other.convert(self.unit).value
        else:
            msg = f"'>=' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __gt__(self, other):
        if not isinstance(other, Unit):
            return self.value > other
        elif type(self) is type(other):
            return self.value > other.convert(self.unit).value
        else:
            msg = f"'>' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def __len__(self):
        try:
            return len(self.value)
        except TypeError:
            return 1

    def __round__(self, n=None):
        return self.kind(round(self.value, n), self.unit)

    def __getitem__(self, item):
        if type(item) is int:
            if item >= len(self):
                raise IndexError
        else:
            raise IndexError
        from .units.unitless import Percentage
        if self.kind is Percentage:
            return self.kind(self.value[item] * 100, self.unit)
        else:
            return self.kind(self.value[item], self.unit)

    def __iter__(self):
        if not hasattr(self.value, '__iter__'):
            return np.array((self.value,)).__iter__()
        else:
            return self.value.__iter__()

    def equals(self, other, precision:int=None):
        if precision is None:
            self.__eq__(other)
        if not isinstance(other, Unit):
            return self.round(precision).__eq__(self.round(precision, value=other))
        elif type(self) is type(other):
            return self.round(precision).__eq__(other.convert(self.unit).round(precision))
        else:
            msg = f"'==' not supported between instances of '{type(self)}' and '{type(other)}'."
            raise TypeError(msg)

    def eq(self, other, precision: int = None):
        return self.equals(other, precision)

    def get_unit(self):
        return self.unit

    def get_units(self):
        return self.get_unit()

    @property
    def units(self):
        return self.unit

    @units.setter
    def units(self, units):
        self.unit = units

    def get_value(self):
        return self.value

    @property
    def values(self):
        return self.value

    @values.setter
    def values(self, value):
        self.value = value

    def round(self, precision:int=0, *, value=None):
        value_ = self.value if value is None else value
        significants = lambda v, i: float(f"{v:.{abs(i)}g}")
        from .units.define import units
        precision = int(precision)
        if precision >= 0 and hasattr(value_, 'round'):
            if _numpy_:
                value_ = value_.round(precision)
            else:
                raise NotImplementedError(f"operations on arrays without Numpy are not implemented.")
        elif precision >= 0:
            value_ = round(value_, precision)
        elif precision < 0 and hasattr(self.value, '__iter__'):
            if _numpy_:
                value_ = np.array([significants(each, precision) for each in value_])
            else:
                raise NotImplementedError(f"operations on arrays without Numpy are not implemented.")
        else:
            value_ = significants(value_, precision)
        return units(value_, self.units) if value is None else value_

    def check_value(self, value):
        if type(value) in (list, tuple):
            try:
                return np.array(value)
            except TypeError:
                raise WrongValueError(str(value))
        elif isinstance(value, Unit):
            return value.value
        elif isinstance(value, Number):
            return value
        elif isinstance(value, _array_like):
            return value
        else:
            raise WrongValueError(str(value))

    def check_unit(self, units):
        if type(units) is str:
            pass
        elif type(self) is units:
            return units.unit
        elif hasattr(units, 'units'):
            units = units.units
        else:
            raise WrongUnitsError(f"'{units}' for '{type(self)}'.")
        if units in self.kind.class_units:
            return units
        else:
            raise WrongUnitsError(f"'{units}' for '{type(self)}'.")



def is_Unit(obj) -> bool:
    return isinstance(obj, Unit)


def valid_unit(unit_name:str) -> bool:
    if unit_name is None or (type(unit_name) is str and unit_name.strip() in Unit._all_units_str):
        return True
    else:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.5.30'
__release__ = 20230724
__all__ = ['units']

from ..dictionaries import dictionary as _dictionary, uncertain_names
from .custom import UserUnits
from .data import *
from .date import *
from .energy import *
from .force import *
from .geometry import *
from .mass import *
from .rates import *
from .ratios import *
from .temperature import *
from .time import *
from .unitless import Dimensionless, Percentage
from ..unit_class import Unit
from ..helpers.common_classes import unit_or_str as unit_or_str, numeric as numeric, numeric_ as _numeric


def units(value: numeric, unit: unit_or_str = None, name=None) -> Unit:
    """
    return an instance of units with the provided value and units.

    Parameters
    ----------
    value : int, float, numeric array, Series or DataFrame
        the value to assign the units.
    unit : str, optional
        the units to assign the value. The default is None.
    name : str, optional
        a string available to user to store metadata.

    Raises
    ------
    TypeError
        if value or units are not the appropriate kind.

    Returns
    -------
    units
        instance of units.

    """
    if unit is None and isinstance(value, Unit):
        value, unit = value.value, value.units
    if unit is None:
        unit = 'Dimensionless'
    if type(unit) is not str:
        raise TypeError("'units' must be a string or Unit instance.")
    if not isinstance(value, _numeric):
        raise TypeError("'value' parameter must be numeric.")

    unit = unit.strip()

    if unit in uncertain_names:
        return Unit(value, unit, name)

    for kind in _dictionary:
        if unit in _dictionary[kind]:
            if "'" in unit:
                u = eval(kind + '''(0, "''' + unit + '''")''', name)
            else:
                u = eval(kind + """(0, '""" + unit + """')""", name)
            if type(u) is Percentage:
                u.value = value / 100
            else:
                u.value = value
            return u

    return UserUnits(value, unit, name)
